__author__ = 'maartenbreddels'
import os
import sys

import numpy as np

import logging
import vaex
import vaex.utils
import vaex.execution
import vaex.file.colfits
import vaex.file.other

max_length = int(1e5)

on_rtd = os.environ.get('READTHEDOCS', None) == 'True'
try:
	import h5py
except:
	if not on_rtd:
		raise
#from vaex.dataset import DatasetLocal

logger = logging.getLogger("vaex.export")

def _export(dataset_input, dataset_output, random_index_column, path, column_names=None, byteorder="=", shuffle=False, selection=False, progress=None, virtual=True, sort=None, ascending=True):
	"""
	:param DatasetLocal dataset: dataset to export
	:param str path: path for file
	:param lis[str] column_names: list of column names to export or None for all columns
	:param str byteorder: = for native, < for little endian and > for big endian
	:param bool shuffle: export rows in random order
	:param bool selection: export selection or not
	:param progress: progress callback that gets a progress fraction as argument and should return True to continue
	:return:
	"""

	if selection:
		if selection == True: # easier to work with the name
			selection = "default"

	N = len(dataset_input) if not selection else dataset_input.selected_length(selection)
	if N == 0:
		raise ValueError("Cannot export empty table")

	if shuffle and sort:
		raise ValueError("Cannot shuffle and sort at the same time")

	if shuffle:
		shuffle_array = dataset_output.columns[random_index_column]


	partial_shuffle = shuffle and len(dataset_input) != len(dataset_input)

	if partial_shuffle:
		# if we only export a portion, we need to create the full length random_index array, and
		shuffle_array_full = np.zeros(len(dataset_input), dtype=byteorder+"i8")
		vaex.vaexfast.shuffled_sequence(shuffle_array_full)
		# then take a section of it
		#shuffle_array[:] = shuffle_array_full[:N]
		shuffle_array[:] = shuffle_array_full[shuffle_array_full<N]
		del shuffle_array_full
		order_array = shuffle_array
	elif shuffle:
		# better to do this in memory
		shuffle_array_memory = np.zeros_like(shuffle_array)
		vaex.vaexfast.shuffled_sequence(shuffle_array_memory)
		shuffle_array[:] = shuffle_array_memory
		order_array = shuffle_array

	if sort:
		if selection:
			raise ValueError("sorting selections not yet supported")
		# these indices sort the input array, but we evaluate the input in sequential order and write it out in sorted order
		# e.g., not b[:] = a[indices]
		# but b[indices_r] = a
		logger.info("sorting...")
		indices = np.argsort(dataset_input.evaluate(sort))
		indices_r = np.zeros_like(indices)
		indices_r[indices] = np.arange(len(indices))
		del indices
		order_array = indices_r if ascending else indices_r[::-1]
		logger.info("sorting done")
		#indices_target

	#i1, i2 = 0, N #len(dataset)
	#print "creating shuffled array"
	if progress == True:
		progress = vaex.utils.progressbar_callable("exporting")
	progress = progress or (lambda value: True)
	progress_total = len(column_names) * len(dataset_input)
	progress_value = 0
	for column_name in column_names:
		logger.debug("  exporting column: %s " % column_name)
		#with vaex.utils.Timer("copying column %s" % column_name, logger):
		if 1:
			block_scope = dataset_input._block_scope(0, vaex.execution.buffer_size_default)
			to_array = dataset_output.columns[column_name]
			if shuffle or sort: # we need to create a in memory copy, otherwise we will do random writes which is VERY inefficient
				to_array_disk = to_array
				if np.ma.isMaskedArray(to_array):
					to_array = np.empty_like(to_array_disk)
				else:
					to_array = np.zeros_like(to_array_disk)
			to_offset = 0 # we need this for selections
			for i1, i2 in vaex.utils.subdivide(len(dataset_input), max_length=max_length):
				logger.debug("from %d to %d (total length: %d, output length: %d)", i1, i2, len(dataset_input), N)
				block_scope.move(i1, i2)
				#	block_scope.move(i1-i1, i2-i1)

				#values = block_scope.evaluate(column_name)
				#if values.dtype.type == np.datetime64:
				#	values = values.view(np.int64)

				if selection:
					mask = dataset_input.evaluate_selection_mask(selection, i1=i1, i2=i2)
					selection_block_length = np.sum(mask)#np.sum(dataset_input.mask[i1:i2])
					if selection_block_length:
						values = block_scope.evaluate(column_name)
						fill_value = np.nan if values.dtype.kind == "f" else None
						#assert np.ma.isMaskedArray(to_array) == np.ma.isMaskedArray(values), "to (%s) and from (%s) array are not of both masked or unmasked (%s)" %\
						#			 (np.ma.isMaskedArray(to_array), np.ma.isMaskedArray(values), column_name)
						if values.dtype.type == np.datetime64:
							values = values.view(np.int64)
						if np.ma.isMaskedArray(to_array) and np.ma.isMaskedArray(values):
							to_array.data[to_offset:to_offset+selection_block_length] = values.filled(fill_value)[mask]
							to_array.mask[to_offset:to_offset+selection_block_length] = values.mask[mask]
						elif not np.ma.isMaskedArray(to_array) and np.ma.isMaskedArray(values):
							to_array[to_offset:to_offset+selection_block_length] = values[mask].filled(fill_value)
						else:
							to_array[to_offset:to_offset+selection_block_length] = values[mask]
						to_offset += selection_block_length
				else:
					values = block_scope.evaluate(column_name)
					fill_value = np.nan if values.dtype.kind == "f" else None
					#assert np.ma.isMaskedArray(to_array) == np.ma.isMaskedArray(values), "to (%s) and from (%s) array are not of both masked or unmasked (%s)" %\
					#			 (np.ma.isMaskedArray(to_array), np.ma.isMaskedArray(values), column_name)
					if values.dtype.type == np.datetime64:
						values = values.view(np.int64)
					if shuffle or sort:
						indices = order_array[i1:i2]
						if np.ma.isMaskedArray(to_array) and np.ma.isMaskedArray(values):
							to_array.data[indices] = values.filled(fill_value)
							to_array.mask[indices] = values.mask
						elif not np.ma.isMaskedArray(to_array) and np.ma.isMaskedArray(values):
							to_array[indices] = values.filled(fill_value)
						else:
							to_array[indices] = values
					else:
						if np.ma.isMaskedArray(to_array) and np.ma.isMaskedArray(values):
							to_array.data[i1:i2] = values.filled(fill_value)
							to_array.mask[i1:i2] = values.mask
						elif np.ma.isMaskedArray(to_array) and np.ma.isMaskedArray(values):
							to_array[i1:i2] = values.filled(fill_value)
						else:
							to_array[i1:i2] = values

				progress_value += i2-i1
				if not progress(progress_value/float(progress_total)):
					break
			if shuffle or sort: # write to disk in one go
				if np.ma.isMaskedArray(to_array) and np.ma.isMaskedArray(to_array_disk):
					to_array_disk.data[:] = to_array.data
					to_array_disk.mask[:] = to_array.mask
				else:
					to_array_disk[:] = to_array
	return column_names

def export_fits(dataset, path, column_names=None, shuffle=False, selection=False, progress=None, virtual=True, sort=None, ascending=True):
	"""
	:param DatasetLocal dataset: dataset to export
	:param str path: path for file
	:param lis[str] column_names: list of column names to export or None for all columns
	:param bool shuffle: export rows in random order
	:param bool selection: export selection or not
	:param progress: progress callback that gets a progress fraction as argument and should return True to continue,
		or a default progress bar when progress=True
	:param: bool virtual: When True, export virtual columns
	:return:
	"""
	if shuffle:
		random_index_name = "random_index"
		while random_index_name in dataset.get_column_names():
			random_index_name += "_new"

	column_names = column_names or dataset.get_column_names(virtual=virtual, strings=True)
	logger.debug("exporting columns(fits): %r" % column_names)
	N = len(dataset) if not selection else dataset.selected_length(selection)
	data_types = []
	data_shapes = []
	ucds = []
	units = []
	for column_name in column_names:
		if column_name in dataset.get_column_names(strings=True):
			column = dataset.columns[column_name]
			shape = (N,) + column.shape[1:]
			dtype = column.dtype
		else:
			dtype = np.float64().dtype
			shape = (N,)
		ucds.append(dataset.ucds.get(column_name))
		units.append(dataset.units.get(column_name))
		data_types.append(dtype)
		data_shapes.append(shape)

	if shuffle:
		column_names.append(random_index_name)
		data_types.append(np.int64().dtype)
		data_shapes.append((N,))
		ucds.append(None)
		units.append(None)
	else:
		random_index_name = None

	# TODO: all expressions can have missing values.. how to support that?
	null_values = {key:dataset.columns[key].fill_value for key in dataset.get_column_names() if dataset.is_masked(key) and dataset.dtype(key).kind != "f"}
	vaex.file.colfits.empty(path, N, column_names, data_types, data_shapes, ucds, units, null_values=null_values)
	if shuffle:
		del column_names[-1]
		del data_types[-1]
		del data_shapes[-1]
	dataset_output = vaex.file.other.FitsBinTable(path, write=True)
	_export(dataset_input=dataset, dataset_output=dataset_output, path=path, random_index_column=random_index_name,
			column_names=column_names, selection=selection, shuffle=shuffle,
			progress=progress, sort=sort, ascending=ascending)
	dataset_output.close_files()

def export_hdf5_v1(dataset, path, column_names=None, byteorder="=", shuffle=False, selection=False, progress=None, virtual=True):
	kwargs = locals()
	import vaex.hdf5.export
	vaex.hdf5.export.export_hdf5_v1(**kwargs)

def export_hdf5(dataset, path, column_names=None, byteorder="=", shuffle=False, selection=False, progress=None, virtual=True, sort=None, ascending=True):
	kwargs = locals()
	import vaex.hdf5.export
	vaex.hdf5.export.export_hdf5(**kwargs)

if __name__ == "__main__":
	sys.exit(main(sys.argv))

def main(argv):
	import argparse
	parser = argparse.ArgumentParser(argv[0])
	parser.add_argument('--verbose', '-v', action='count', default=0)
	parser.add_argument('--quiet', '-q', default=False, action='store_true', help="do not output anything")
	parser.add_argument('--list', '-l', default=False, action='store_true', help="list columns of input")
	parser.add_argument('--progress', help="show progress (default: %(default)s)", default=True, action='store_true')
	parser.add_argument('--no-progress', dest="progress", action='store_false')
	parser.add_argument('--shuffle', "-s", dest="shuffle", action='store_true', default=False)
	parser.add_argument('--sort', dest="sort", default=None)
	parser.add_argument('--virtual', dest="virtual", action='store_true', default=False, help="Also export virtual columns")
	parser.add_argument('--fraction', "-f", dest="fraction", type=float, default=1.0, help="fraction of input dataset to export")

	subparsers = parser.add_subparsers(help='type of input source', dest="task")

	parser_soneira = subparsers.add_parser('soneira', help='create soneira peebles dataset')
	parser_soneira.add_argument('output', help='output file')
	parser_soneira.add_argument("columns", help="list of columns to export (or all when empty)", nargs="*")
	parser_soneira.add_argument('--dimension','-d', type=int, help='dimensions', default=4)
	#parser_soneira.add_argument('--eta','-e', type=int, help='dimensions', default=3)
	parser_soneira.add_argument('--max-level','-m', type=int, help='dimensions', default=28)
	parser_soneira.add_argument('--lambdas','-l', type=int, help='lambda values for fractal', default=[1.1, 1.3, 1.6, 2.])


	parser_tap = subparsers.add_parser('tap', help='use TAP (Table Access Protocol) as source')
	parser_tap.add_argument("tap_url", help="input source or file")
	parser_tap.add_argument("table_name", help="input source or file")
	parser_tap.add_argument("output", help="output file (ends in .fits or .hdf5)")
	parser_tap.add_argument("columns", help="list of columns to export (or all when empty)", nargs="*")

	parser_file = subparsers.add_parser('file', help='use a file as source (e.g. .hdf5, .fits, .vot (VO table), .asc (ascii)')
	parser_file.add_argument("input", help="input source or file, when prefixed with @ it is assumed to be a text file with a file list (one file per line)")
	parser_file.add_argument("output", help="output file (ends in .fits or .hdf5)")
	parser_file.add_argument("columns", help="list of columns to export (or all when empty)", nargs="*")

	parser_file = subparsers.add_parser('csv', help='use a csv file as source (e.g. .hdf5, .fits, .vot (VO table), .asc (ascii)')
	parser_file.add_argument("input", help="input source or file, when prefixed with @ it is assumed to be a text file with a file list (one file per line)")
	parser_file.add_argument("output", help="output file (ends in .hdf5)")
	parser_file.add_argument("columns", help="list of columns to export (or all when empty)", nargs="*")

	args = parser.parse_args(argv[1:])

	verbosity = ["ERROR", "WARNING", "INFO", "DEBUG"]
	logging.getLogger("vaex").setLevel(verbosity[min(3, args.verbose)])
	dataset = None
	if args.task == "soneira":
		if vaex.utils.check_memory_usage(4*8*2**args.max_level, vaex.utils.confirm_on_console):
			if not args.quiet:
				print("generating soneira peebles dataset...")
			dataset = vaex.file.other.SoneiraPeebles(args.dimension, 2, args.max_level, args.lambdas)
		else:
			return 1
	if args.task == "tap":
		dataset = vaex.dataset.DatasetTap(args.tap_url, args.table_name)
		if not args.quiet:
			print("exporting from {tap_url} table name {table_name} to {output}".format(tap_url=args.tap_url, table_name=args.table_name, output=args.output))
	if args.task == "csv":
		#dataset = vaex.dataset.DatasetTap(args.tap_url, args.table_name)
		if not args.quiet:
			print("exporting from {input} to {output}".format(input=args.input, output=args.output))
	if args.task == "file":
		if args.input[0] == "@":
			inputs = open(args.input[1:]).readlines()
			dataset = vaex.open_many(inputs)
		else:
			dataset = vaex.open(args.input)
		if not args.quiet:
			print("exporting from {input} to {output}".format(input=args.input, output=args.output))

	if dataset is None and args.task not in ["csv"]:
		if not args.quiet:
			print("Cannot open input")
		return 1
	if dataset:
		dataset.set_active_fraction(args.fraction)
	if args.list:
		if not args.quiet:
			print("columns names: " + " ".join(dataset.get_column_names()))
	else:
		if args.task == "csv":
			row_count = -1 # the header does not count
			with file(args.input) as lines:
				for line in lines:
					row_count += 1
					#print line
			logger.debug("row_count: %d", row_count)
			with file(args.input) as lines:
				line = next(lines).strip()
				#print line
				names = line.strip().split(",")
				line = next(lines).strip()
				values = line.strip().split(",")
				numerics = []
				for value in values:
					try:
						float(value)
						numerics.append(True)
					except:
						numerics.append(False)
				names_numeric = [name for name, numeric in zip(names, numerics) if numeric]
				print (names_numeric)
				output = vaex.file.other.Hdf5MemoryMapped.create(args.output, row_count, names_numeric)
				Ncols = len(names)
				cols = [output.columns[name] if numeric else None for name, numeric in zip(names, numerics)]
				def copy(line, row_index):
					values = line.strip().split(",")
					for column_index in range(Ncols):
						if numerics[column_index]:
							value = float(values[column_index])
							cols[column_index][row_index] = value
				row = 0
				copy(line, row)
				row += 1
				progressbar = vaex.utils.progressbar("exporting") if args.progress else None
				for line in lines:
					#print line
					copy(line, row)
					row += 1
					if row % 1000:
						progressbar.update(row/float(row_count))
				progressbar.finish()
				#print names
		else:
			if args.columns:
				columns = args.columns
			else:
				columns = None
			if columns is None:
				columns = dataset.get_column_names(strings=True, virtual=args.virtual)
			for column in columns:
				if column not in dataset.get_column_names(strings=True, virtual=True):
					if not args.quiet:
						print("column %r does not exist, run with --list or -l to list all columns" % column)
					return 1

			base, output_ext = os.path.splitext(args.output)
			if output_ext not in [".hdf5", ".fits"]:
				if not args.quiet:
					print("extension %s not supported, only .fits and .hdf5 are" % output_ext)
				return 1

			if not args.quiet:
				print("exporting %d rows and %d columns" % (len(dataset), len(columns)))
				print("columns: " +" ".join(columns))
			progressbar = vaex.utils.progressbar("exporting") if args.progress else None

			def update(p):
				if progressbar:
					progressbar.update(p)
				return True
			if output_ext == ".hdf5":
				export_hdf5(dataset, args.output, column_names=columns, progress=update, shuffle=args.shuffle, sort=args.sort)
			elif output_ext == ".fits":
				export_fits(dataset, args.output, column_names=columns, progress=update, shuffle=args.shuffle, sort=args.sort)
			if progressbar:
				progressbar.finish()
			if not args.quiet:
				print("\noutput to %s" % os.path.abspath(args.output))
			dataset.close_files()
	return 0



# -*- coding: utf-8 -*-
import time
import platform
import os
import sys
import numpy as np
import math
import concurrent.futures
import six
import re

is_frozen = getattr(sys, 'frozen', False)
PY2 = sys.version_info[0] == 2
PY3 = sys.version_info[0] == 3

osname = dict(darwin="osx", linux="linux", windows="windows")[platform.system().lower()]

class AttrDict(dict):
	def __init__(self, *args, **kwargs):
		dict.__init__(self, *args, **kwargs)

	def __getattr__(self, name):
		return self[name]

	def __setattr__(self, key, value):
		self[key] = value

def subdivide(length, parts=None, max_length=None):
	"""Generates a list with start end stop indices of length parts, [(0, length/parts), ..., (.., length)]"""
	if max_length:
		i1 = 0
		done = False
		while not done:
			i2 = min(length, i1 + max_length)
			#print i1, i2
			yield i1, i2
			i1 = i2
			if i1 == length:
				done = True
	else:
		part_length = int(math.ceil(float(length)/parts))
		#subblock_count = math.ceil(total_length/subblock_size)
		#args_list = []
		for index in range(parts):
			i1, i2 = index * part_length, min(length, (index +1) * part_length)
			yield i1, i2


def submit_subdivide(thread_count, f, length, max_length):
	futures = []
	thread_pool = concurrent.futures.ThreadPoolExecutor(thread_count)
	#thread_pool = concurrent.futures.ProcessPoolExecutor(thread_count)
	for i1, i2 in list(subdivide(length, max_length=max_length)):
		futures.append(thread_pool.submit(f, i1, i2))
	return futures

def linspace_centers(start, stop, N):
	return np.arange(N) / (N+0.) * (stop-start) + float(stop-start)/N/2 + start

def multisum(a, axes):
	correction = 0
	for axis in axes:
		a = np.nansum(a, axis=axis-correction)
		correction += 1
	return a

def disjoined(data):
	# create marginalized distributions and multiple them together
	data_disjoined = None
	dim = len(data.shape)
	for d in range(dim):
		axes = list(range(dim))
		axes.remove(d)
		data1d = multisum(data, axes)
		shape = [1 for k in range(dim)]
		shape[d] = len(data1d)
		data1d = data1d.reshape(tuple(shape))
		if d == 0:
			data_disjoined = data1d
		else:
			data_disjoined = data_disjoined * data1d
	return data_disjoined



def get_data_file(filename):
	try: # this works for egg like stuff, but fails for py2app apps
		from pkg_resources import Requirement, resource_filename
		path = resource_filename(Requirement.parse("vaex"), filename)
		if os.path.exists(path):
			return path
	except:
		pass
	# this is where we expect data to be in normal installations
	for extra in ["", "data", "data/dist", "../data", "../data/dist"]:
		path = os.path.join(os.path.dirname(__file__), "..", "..", extra, filename)
		if os.path.exists(path):
			return path
		path = os.path.join(sys.prefix, extra, filename)
		if os.path.exists(path):
			return path
		# if all fails..
		path = os.path.join(get_root_path(), extra, filename)
		if os.path.exists(path):
			return path

def get_root_path():
	osname = platform.system().lower()
	#if (osname == "linux") and is_frozen: # we are using pyinstaller
	if is_frozen: # we are using pyinstaller or py2app
		return os.path.dirname(sys.argv[0])
	else:
		return os.path.abspath(".")
	
	
def os_open(document):
	"""Open document by the default handler of the OS, could be a url opened by a browser, a text file by an editor etc"""
	osname = platform.system().lower()
	if osname == "darwin":
		os.system("open \"" +document +"\"")
	if osname == "linux":
		cmd = "xdg-open \"" +document +"\"&"
		os.system(cmd)
	if osname == "windows":
		os.system("start \"" +document +"\"")

def filesize_format(value):
	for unit in ['bytes','KiB','MiB','GiB']:
		if value < 1024.0:
			return "%3.1f%s" % (value, unit)
		value /= 1024.0
	return "%3.1f%s" % (value, 'TiB')



log_timer = True
class Timer(object):
	def __init__(self, name=None, logger=None):
		self.name = name
		self.logger = logger

	def __enter__(self):
		if log_timer:
			if self.logger:
				self.logger.debug("%s starting" % self.name)
			else:
				print(('[%s starting]...' % self.name))
			self.tstart = time.time()

	def __exit__(self, type, value, traceback):
		if log_timer:
			msg = "%s done, %ss elapsed" % (self.name, time.time() - self.tstart)
			if self.logger:
				self.logger.debug(msg)
			else:
				print(msg)
			if type or value or traceback:
				print((type, value, traceback))
		return False

def get_private_dir(subdir=None):
	path = os.path.expanduser('~/.vaex')
	if subdir:
		path = os.path.join(path, subdir)
	if not os.path.exists(path):
		os.makedirs(path)
	return path

def make_list(sequence):
	if isinstance(sequence, np.ndarray):
		return sequence.tolist()
	else:
		return list(sequence)

import progressbar
#from progressbar import AnimatedMarker, Bar, BouncingBar, Counter, ETA, \
#   FileTransferSpeed, FormatLabel, Percentage, \
#    ProgressBar, ReverseBar, RotatingMarker, \
#    SimpleProgress, Timer, AdaptiveETA, AbsoluteETA, AdaptiveTransferSpeed
#from progressbar.widgets import TimeSensitiveWidgetBase, FormatWidgetMixin

class CpuUsage(progressbar.widgets.FormatWidgetMixin, progressbar.widgets.TimeSensitiveWidgetBase):
	def __init__(self, format='CPU Usage: %(cpu_usage)s%%', usage_format="% 5d"):
		super(CpuUsage, self).__init__(format=format)
		self.usage_format = usage_format
		self.utime_0 = None
		self.stime_0 = None
		self.walltime_0 = None

	def __call__(self, progress, data):
		utime, stime, child_utime, child_stime, walltime = os.times()
		if self.utime_0 is None:
			self.utime_0 = utime
		if self.stime_0 is None:
			self.stime_0 = stime
		if self.walltime_0 is None:
			self.walltime_0 = walltime
		data["utime_0"] = self.utime_0
		data["stime_0"] = self.stime_0
		data["walltime_0"] = self.walltime_0

		delta_time = utime - self.utime_0 + stime - self.stime_0
		delta_walltime = walltime - self.walltime_0
		#print delta_time, delta_walltime, utime, self.utime_0, stime, self.stime_0
		if delta_walltime == 0:
			data["cpu_usage"] = "---"
		else:
			cpu_usage = delta_time/(delta_walltime * 1.) * 100
			data["cpu_usage"] = self.usage_format % cpu_usage
		#utime0, stime0, child_utime0, child_stime0, walltime0 = os.times()
		return progressbar_mod.widgets.FormatWidgetMixin.__call__(self, progress, data)
progressbar_mod = progressbar
def _progressbar_progressbar2(type=None, name="processing", max_value=1):
	widgets = [
		name,
        ': ', progressbar_mod.widgets.Percentage(),
        ' ', progressbar_mod.widgets.Bar(),
        ' ', progressbar_mod.widgets.ETA(),
        #' ', progressbar_mod.widgets.AdaptiveETA(),
		' ', CpuUsage()
    ]
	bar = progressbar_mod.ProgressBar(widgets=widgets, max_value=max_value)
	bar.start()
	return bar
	#FormatLabel('Processed: %(value)d lines (in: %(elapsed)s)')

def _progressbar_vaex(type=None, name="processing", max_value=1):
	import vaex.misc.progressbar as pb
	return pb.ProgressBar(0, 1)


class _ProgressBarWidget(object):
	def __init__(self, name=None):
		import ipywidgets as widgets
		from IPython.display import display
		self.value = 0
		self.widget = widgets.FloatProgress(min=0, max=1)
		self.widget.description = name or 'Progress:'
		display(self.widget)

	def __call__(self, value):
		self.value = value
		self.widget.value = value

	def update(self, value):
		self(value)

	def finish(self):
		self.widget.description = 'Finished:'
		self(1)

	# def add_child(self, progress, task, name):
	# 	#print('add', name)
	# 	pb = _ProgressBarWidget(name=name)
	# 	task.signal_progress.connect(pb)

_progressbar_typemap = {}
_progressbar_typemap['progressbar2'] = _progressbar_progressbar2
_progressbar_typemap['vaex'] = _progressbar_vaex
_progressbar_typemap['widget'] = _ProgressBarWidget

def progressbar(type_name=None, name="processing", max_value=1):
	type_name = type_name or 'vaex'
	return _progressbar_typemap[type_name](name=name)

def progressbar_widget():
	pass
class _progressbar(object):
	pass
class _progressbar_wrapper(_progressbar):
	def __init__(self, bar):
		self.bar = bar

	def __call__(self, fraction):
		self.bar.update(fraction)
		if fraction == 1:
			self.bar.finish()
		return True

	def status(self, name):
		self.bar.bla = name

class _progressbar_wrapper_sum(_progressbar):
	def __init__(self, children=None, next=None, bar=None, parent=None, name=None):
		self.next = next
		self.children = children or list()
		self.finished = False
		self.last_fraction = None
		self.fraction = 0
		self.bar = bar
		self.parent = parent
		self.name = name
		self.cancelled = False
		self.oncancel = lambda : None

	def cancel(self):
		self.cancelled = True

	def __repr__(self):
		name = self.__class__.__module__ + "." +self.__class__.__name__
		return "<%s(name=%r)> instance at 0x%x" % (name, self.name, id(self))

	def add(self, name=None):
		pb = _progressbar_wrapper_sum(parent=self, name=name)
		self.children.append(pb)
		return pb

	def add_task(self, task, name=None):
		pb = self.add(name)
		pb.oncancel = task.cancel
		task.signal_progress.connect(pb)
		if self.bar and hasattr(self.bar, 'add_child'):
			self.bar.add_child(pb, task, name)

	def __call__(self, fraction):
		if self.cancelled:
			return False
		# ignore fraction
		result = True
		if len(self.children) == 0:
			self.fraction = fraction
		else:
			self.fraction = sum([c.fraction for c in self.children])/len(self.children)
		fraction = self.fraction
		if fraction != self.last_fraction: # avoid too many calls
			if fraction == 1 and not self.finished: # make sure we call finish only once
				self.finished = True
				if self.bar:
					self.bar.finish()
			elif fraction != 1:
				if self.bar:
					self.bar.update(fraction)
			if self.next:
				result = self.next(fraction)
		if self.parent:
			assert self in self.parent.children
			result = self.parent(None) in [None, True] and result # fraction is not used anyway..
			if result is False:
				self.oncancel()
		self.last_fraction = fraction
		return result

	def status(self, name):
		pass

def progressbars(f=True,next=None, name=None):
	if isinstance(f, _progressbar_wrapper_sum):
		return f
	if callable(f):
		next = f
		f = False
	if f in [None, False]:
		return _progressbar_wrapper_sum(next=next, name=name)
	else:
		if f is True:
			return _progressbar_wrapper_sum(bar=progressbar(), next=next, name=name)
		elif isinstance(f, six.string_types):
			return _progressbar_wrapper_sum(bar=progressbar(f), next=next, name=name)
		else:
			return _progressbar_wrapper_sum(next=next, name=name)


def progressbar_callable(name="processing", max_value=1):
	bar = progressbar(name, max_value=max_value)
	return _progressbar_wrapper(bar)




def confirm_on_console(topic, msg):
	done = False
	print(topic)
	while not done:
		output = raw_input(msg +":[y/n]")
		if output.lower() == "y":
			return True
		if output.lower() == "n":
			return False

import json
import yaml

def yaml_dump(f, data):
	yaml.safe_dump(data, f, default_flow_style=False, encoding='utf-8',  allow_unicode=True)
def yaml_load(f):
	return yaml.safe_load(f)

def write_json_or_yaml(filename, data):
	base, ext = os.path.splitext(filename)
	if ext == ".json":
		with open(filename, "w") as f:
			json.dump(data, f, indent=2)
	elif ext == ".yaml":
		with open(filename, "w") as f:
			yaml_dump(f, data)
	else:
		raise ValueError("file should end in .json or .yaml (not %s)" % ext)

def read_json_or_yaml(filename):
	base, ext = os.path.splitext(filename)
	if ext == ".json":
		with open(filename, "r") as f:
			return json.load(f) or {}
	elif ext == ".yaml":
		with open(filename, "r") as f:
			return yaml_load(f) or {}
	else:
		raise ValueError("file should end in .json or .yaml (not %s)" % ext)

# from http://stackoverflow.com/questions/5121931/in-python-how-can-you-load-yaml-mappings-as-ordereddicts
import collections
_mapping_tag = yaml.resolver.BaseResolver.DEFAULT_MAPPING_TAG

def dict_representer(dumper, data):
    return dumper.represent_dict(data.iteritems() if hasattr(data, "iteritems") else data.items())

def dict_constructor(loader, node):
    return collections.OrderedDict(loader.construct_pairs(node))

yaml.add_representer(collections.OrderedDict, dict_representer, yaml.SafeDumper)
yaml.add_constructor(_mapping_tag, dict_constructor, yaml.SafeLoader)

import psutil
def check_memory_usage(bytes_needed, confirm):
	if bytes_needed > psutil.virtual_memory().available:
		if bytes_needed < (psutil.virtual_memory().available +psutil.swap_memory().free):
			text = "Action requires %s, you have enough swap memory available but it will make your computer slower, do you want to continue?" % (filesize_format(bytes_needed),)
			return confirm("Memory usage issue", text)
		else:
			text = "Action requires %s, you do not have enough swap memory available, do you want try anyway?" % (filesize_format(bytes_needed),)
			return confirm("Memory usage issue", text)
	return True

import six
def ensure_string(string_or_bytes, encoding="utf-8", cast=False):
	if cast:
		if six.PY2:
			string_or_bytes = unicode(string_or_bytes)
		else:
			string_or_bytes = str(string_or_bytes)
	if isinstance(string_or_bytes, six.string_types):
		return string_or_bytes
	else:
		return string_or_bytes.decode(encoding)

def filename_shorten(path, max_length=150):
	#parts = path.split(os.path.sep)
	parts = []
	done = False
	tail = path
	while not done:
		tail, head = os.path.split(tail)
		#print ">>", tail, head
		if not head:
			done = True
			parts.append(tail)
		else:
			parts.append(head)
	parts.reverse()
	#print "parts", parts, os.path.join(*parts)
	if len(parts) > 4:
		first, middle, last = os.path.join(parts[0], parts[1]), parts[2:-1], parts[-1]
		#print first, middle, last
		while (len(os.path.join(first, *(middle + [last]))) <= max_length) and middle:
			last = os.path.join(middle[-1], last)
			middle = middle[:-1]
		if middle:
			return os.path.join(first, "...", last)
		else:
			return os.path.join(first, last)
	else:
		return path

def listify(*args):
	if isinstance(args[0], six.string_types):
		return False, [[x] for x in args]
	try:
		_ = args[0][0]
		return True, args
	except:
		return False, [[x] for x in args]

def unlistify(waslist, *args):
	if waslist:
		if len(args) == 1:
			return args[0]
		return args
	else:
		values =  [x[0] for x in args]
		if len(values) == 1:
			return values[0]

def find_valid_name(name, used=[]):
	first, rest = name[0], name[1:]
	name = re.sub("[^a-zA-Z_]", "_", first) +  re.sub("[^a-zA-Z_0-9]", "_", rest)
	if name in used:
		nr = 1
		while name + ("_%d" % nr) in used:
			nr += 1
		name = name + ("_%d" % nr)
	return name
_python_save_name = find_valid_name

import contextlib
@contextlib.contextmanager
def write_to(f, mode):
    """Flexible writing, where f can be a filename or f object, if filename, closed after writing"""
    if hasattr(f, 'write'):
        yield f
    else:
        f = open(f, mode)
        yield f
        f.close()


# these are used to add namespace to vaex' dataset
# like ds.viz.plot2d(ds.x, ds.y)

class BoundMethods(object):

    def __init__(self, obj, methods):
        for name, value in methods.items():
            setattr(self, name, value.__get__(obj))


class InnerNamespace(object):

    def __init__(self, methods):
        self._methods = methods

    def _add(self, **kwargs):
        self._methods.update(kwargs)
        self.__dict__.update(kwargs)

    # def __getattr__(self, name):
    # 	if name in self._methods:
    # 		return self._methods[name]
    # 	else:
    # 		return object.__getattr__(self, name)

    def __get__(self, obj, objtype):
        #print("get", obj, objtype)
        if obj is None:
            return self
        else:
            return BoundMethods(obj, self._methods)



import sys, os, imp
from setuptools import Extension

dirname = os.path.dirname(__file__)
path_version = os.path.join(dirname, "../vaex/version.py")
version = imp.load_source('version', path_version)

name        = 'vaex'
author      = "Maarten A. Breddels",
author_email= "maartenbreddels@gmail.com",
license     = 'MIT'
version     = version.versionstring
url         = 'https://www.github.com/maartenbreddels/vaex'
# TODO: can we do without requests and progressbar2?
# TODO: after python2 supports frops, future and futures can also be dropped
# TODO: would be nice to have astropy only as dep in vaex-astro
install_requires_core = ["numpy>=1.11", "scipy", "astropy>=1", "aplus", "futures>=2.2.0",
    "future>=0.15.2", "pyyaml", "progressbar2", "psutil>=1.2.1", "requests", "six"]
install_requires_viz = ["matplotlib>=1.3.1", ]
install_requires_server = ["tornado>4.1", "cachetools"]
install_requires_astro = ["kapteyn"]

packages = ["vaex"]

class get_numpy_include(object):
    """Helper class to determine the numpy include path
    The purpose of this class is to postpone importing numpy
    until it is actually installed, so that the ``get_include()``
    method can be invoked. """

    def __init__(self):
        pass

    def __str__(self):
        import numpy as np
        return np.get_include()

extra_compile_args = ["-std=c++0x", "-mfpmath=sse", "-O3", "-funroll-loops"]

extension_vaexfast = Extension("vaex.vaexfast", [os.path.join(dirname, "../src/vaex/vaexfast.cpp")],
                include_dirs=[get_numpy_include()],
                extra_compile_args=extra_compile_args)



__author__ = 'maartenbreddels'
import unittest
import os
import tempfile
import logging
import numpy as np
import PIL.Image
import PIL.ImageChops

import vaex as vx
import vaex.ui
import vaex.ui.main
import vaex.ui.layers
import vaex.utils
import vaex.dataset
import vaex.execution
import vaex.webserver
from vaex.ui.qt import QtGui, QtCore, QtTest

import vaex.ui.qt as dialogs
import random

# this will trigger more code, such as the canceling in between computation
vaex.execution.buffer_size = 10000
#vx.set_log_level_off()

example_path = vaex.utils.get_data_file("helmi-dezeeuw-2000-10p.hdf5")
vaex.ui.hidden = True

qt_app = QtGui.QApplication([])

base_path = os.path.dirname(__file__)
def get_comparison_image(name):
	osname = vaex.utils.osname
	return os.path.join(base_path, "images", name+"_" + osname + ".png")

#logging.getLogger("vaex.ui.queue").setLevel(logging.DEBUG)
#logging.getLogger("vaex.ui").setLevel(logging.DEBUG)
vx.set_log_level_warning()
#vx.set_log_level_debug()
import logging
logger = logging.getLogger("vaex.test.ui")


overwrite_images = False

class CallCounter(object):
	def __init__(self, return_value=None):
		self.counter = 0
		self.return_value = return_value

	def __call__(self, *args, **kwargs):
		self.counter += 1
		return self.return_value

class TestApp(unittest.TestCase):
	def setUp(self):
		self.dataset = vaex.dataset.DatasetArrays("dataset")

		self.x = x = np.arange(10)
		self.y = y = x ** 2
		self.dataset.add_column("x", x)
		self.dataset.add_column("y", y)
		self.dataset.set_variable("t", 1.)
		self.dataset.add_virtual_column("z", "x+t*y")

		self.app = vx.ui.main.VaexApp()

	def test_table(self):
		ds = vx.example()
		self.app.dataset_selector.add(ds)
		self.dataset.set_current_row(3)
		table = self.app.dataset_panel.tableview()
		self.dataset.set_current_row(0)
		model = table.tableModel.createIndex(1, 1)
		table.tableView.pressed.emit(model)
		self.assertEqual(ds.get_current_row(), 1)
		QtTest.QTest.qWait(100) # make sure it gets rendered (is this the good way?)
		#QtTest.QTest.mouseClick(table.count_from_zero, QtCore.Qt.LeftButton)
		table.count_from_zero.setCheckState(QtCore.Qt.Checked)
		QtTest.QTest.qWait(10000) # make sure it gets rendered (is this the good way?)

	def test_generate_data(self):
		actions = self.app.menu_data.actions()
		for action in actions:
			if "Soneira Peebles" in action.text():
				action.trigger()
				self.assertIsInstance(self.app.current_dataset, vx.file.other.SoneiraPeebles)
				break

		for action in actions:
			if "Zeldovich" in action.text():
				action.trigger()
				self.assertIsInstance(self.app.current_dataset, vx.file.other.Zeldovich)
				break

	def test_default(self):
		app = vx.ui.main.VaexApp(open_default=False)
		self.assert_(app.dataset_selector.is_empty())
		self.assertEqual(None, app.current_dataset)

		app = vx.ui.main.VaexApp(open_default=True)
		self.assert_(not app.dataset_selector.is_empty())
		self.assertIsNotNone(app.current_dataset)

	def test_add_dataset(self):
		app = vx.ui.main.VaexApp()
		ds = vx.example()
		app.dataset_selector.add(ds)
		self.assert_(not app.dataset_selector.is_empty())
		self.assertEqual(int(app.dataset_panel.label_length.text().replace(",", "")), len(ds))
		self.assertEqual(ds, app.current_dataset)

	def test_open_dataset(self):
		app = vx.ui.main.VaexApp()
		ds = app.dataset_selector.open(example_path)
		self.assert_(not app.dataset_selector.is_empty())
		self.assertEqual(int(app.dataset_panel.label_length.text().replace(",", "")), len(ds))
		self.assertEqual(ds, app.current_dataset)

	def test_export(self):
		path_hdf5 = tempfile.mktemp(".hdf5")
		path_hdf5_ui = tempfile.mktemp(".hdf5")
		path_fits = tempfile.mktemp(".fits")
		path_fits_ui = tempfile.mktemp(".fits")

		for dataset in [self.dataset]:
			self.app.dataset_selector.add(dataset)
			for fraction in [1, 0.5]:
				dataset.set_active_fraction(fraction)
				dataset.select("x > 3")
				length = len(dataset)
				# TODO: gui doesn't export virtual columns, add "z" to this list
				for column_names in [["x", "y"], ["x"], ["y"]]:
					for byteorder in "=<>":
						for shuffle in [False, True]:
							for selection in [False, True]:
								for export in [dataset.export_fits, dataset.export_hdf5] if byteorder == ">" else [dataset.export_hdf5]:
									type = "hdf5" if export == dataset.export_hdf5 else "fits"
									if shuffle and selection:
										continue # TODO: export should fail on this combination
									#print column_names, byteorder, shuffle, selection, type
									if export == dataset.export_hdf5:
										path = path_hdf5
										path_ui = path_hdf5_ui
										export(path, column_names=column_names, byteorder=byteorder, shuffle=shuffle, selection=selection)
									else:
										path = path_fits
										path_ui = path_fits_ui
										export(path, column_names=column_names, shuffle=shuffle, selection=selection)
									compare_direct = vx.open(path)

									dialogs.set_choose(1 if selection else 0).then("=<>".index(byteorder))
									# select columns
									dialogs.set_select_many(True, [name in column_names for name in dataset.get_column_names()])
									counter_confirm = CallCounter(return_value=shuffle)
									counter_info = CallCounter()
									dialogs.dialog_confirm = counter_confirm
									dialogs.dialog_info = counter_info
									dialogs.get_path_save = lambda *args: path_ui
									dialogs.ProgressExecution = dialogs.FakeProgressExecution
									import sys
									sys.stdout.flush()

									self.app.export(type=type)
									compare_ui = vx.open(path_ui)

									column_names = column_names or ["x", "y", "z"]
									self.assertEqual(compare_direct.get_column_names(), compare_ui.get_column_names())
									for column_name in column_names:
										values_ui = compare_ui.evaluate(column_name)
										values = compare_direct.evaluate(column_name)
										self.assertEqual(sorted(values), sorted(values_ui))


class __TestPlotPanel(unittest.TestCase):
	def setUp(self):
		self.app = vx.ui.main.VaexApp([], open_default=True)

	def test_open_and_close(self):
		button = self.app.dataset_panel.button_2d
		self.app.show()
		self.app.hide()
		self.assert_(len(self.app.windows) == 0)
		QtTest.QTest.mouseClick(button, QtCore.Qt.LeftButton)
		self.assertEqual(len(self.app.windows), 1)
		self.assertEqual(self.app.windows[0], self.app.current_window)
		self.assertEqual(self.app.windows[0].dataset, self.app.current_dataset)

		self.app.current_window.close()
		self.assert_(len(self.app.windows) == 0)
		self.assertEqual(None, self.app.current_window)

from vaex.ui.plot_windows import PlotDialog

# hide the class from the main namespace
class NoTest:
	class TestPlotPanel(unittest.TestCase):
		def create_app(self):
			self.app = vx.ui.main.VaexApp([], open_default=True)
		def setUp(self):
			vaex.dataset.main_executor = None
			vaex.dataset.main_executor = None
			vaex.multithreading.main_pool = None
			self.create_app()
			self.app.show()
			self.app.hide()
			self.open_window()
			self.window = self.app.current_window
			#self.window.xlabel = ""
			#self.window.ylabel = ""
			self.window.set_plot_size(512, 512)
			self.window.show()
			self.window.hide()
			self.layer = self.window.current_layer
			self.layer.state.colorbar = False
			self.no_exceptions = True
			import sys
			def testExceptionHook(type, value, tback):
				self.no_exceptions = False
				self.exception_info = (type, value, tback)
				sys.__excepthook__(type, value, tback)

			sys.excepthook = testExceptionHook

			self.no_error_in_field = True
			def error_in_field(*args):
				print(args)
				self.no_error_in_field = False
				previous_error_in_field(*args)
			previous_error_in_field = vaex.ui.layers.LayerTable.error_in_field
			vaex.ui.layers.LayerTable.error_in_field = error_in_field
			def log_error(*args):
				print("dialog error", args)
			dialogs.dialog_error = log_error

		def test_favorite_selections(self):
			with dialogs.assertError():
				self.window.action_selection_add_favorites.trigger()
			self.window.current_layer.dataset.select("x > 5")
			name = "test-selection-" + str(random.random())
			with dialogs.settext(name):
				self.window.action_selection_add_favorites.trigger()
				self.assertIn(name, self.window.current_layer.dataset.favorite_selections)
				found = 0
				found_index = 0
				for i, action in enumerate(self.window._favorite_selections_actions):
					if action.text() == name:
						found += 1
						index = i
				self.assertEqual(found, 1, "expect the entry to occur once, occurred %d" % found)
			with dialogs.setchoose(index):
				self.window.action_selection_remove_favorites.trigger()
				self.assertNotIn(name, self.window.current_layer.dataset.favorite_selections)
				found = 0
				for action in self.window.menu_selection.actions():
					if action.text() == name:
						found += 1
				self.assertEqual(found, 0, "expect the entry to occur be removed, but it occurred %d" % found)
			self.window.dataset.favorite_selections.clear()
			with dialogs.assertError():
				self.window.action_selection_remove_favorites.trigger()
			self.window.remove_layer()
			with dialogs.assertError():
				self.window.action_selection_remove_favorites.trigger()
			with dialogs.assertError():
				self.window.action_selection_add_favorites.trigger()

		#def test_bla(self):
		#	self.window.current_layer.x = "dsadsa"
		#	self.window._wait()
		def tearDown(self):
			#vx.promise.rereaise_unhandled()
			for dataset in self.app.dataset_selector.datasets:
				dataset.close_files()
				dataset.remove_virtual_meta()
			self.window.close()
			if not self.no_exceptions:
				type, value, tback = self.exception_info
				import traceback
				print("printing traceback:")
				traceback.print_exception(type, value, tback)

			self.assertTrue(self.no_exceptions)
			self.assertTrue(self.no_error_in_field)

		def compare(self, fn1, fn2):
			if not os.path.exists(fn2) and overwrite_images:
				import shutil
				shutil.copy(fn1, fn2)
			assert os.path.exists(fn2), "image missing: cp {im1} {im2}".format(im1=fn1, im2=fn2)

			try:
				image1 = PIL.Image.open(fn1)
				image2 = PIL.Image.open(fn2)
				diff = PIL.ImageChops.difference(image1, image2)
				extrema = diff.getextrema()
				for i, (vmin, vmax) in enumerate(extrema):
					msg = "difference found between {im1} and {im2} in band {band}\n $ cp {im1} {im2}".format(im1=fn1, im2=fn2, band=i)
					if vmin != vmax and overwrite_images:
						image1.show()
						image2.show()
						done = False
						while not done:
							answer = raw_input("is the new image ok? [y/N]").lower().strip()
							if answer == "n":
								self.assertEqual(vmin, 0, msg)
								return
							if answer == "y":
								import shutil
								shutil.copy(fn1, fn2)
								return
					self.assertEqual(vmin, 0, msg)
					self.assertEqual(vmax, 0, msg)
			finally:
				pass
				#image1.close()
				#image2.close()

		def test_navigation_history(self):
			self.window._wait()
			self._assert_default_image()

			self._do_zoom()
			self.window._wait()
			self._assert_zoom_image()

			self.window.action_undo.trigger()
			self.window._wait()
			self._assert_default_image()

			self.window.action_redo.trigger()
			self.window._wait()
			self._assert_zoom_image()

		def test_empty_region(self):
			self._move_to_empty_region()
			self._assert_empty_image()
			self.window._wait()


class TestPlotPanel1d(NoTest.TestPlotPanel):
	def _move_to_empty_region(self):
		x = 1e4
		self.window.current_layer.xlim = [-10+x, 10+x]

	def _do_zoom(self):
		self.window.current_layer.xlim = [-10, 10]

	def _assert_default_image(self):
		filename = self.window.plot_to_png()
		self.compare(filename, get_comparison_image("default1d"))

	def _assert_empty_image(self):
		filename = self.window.plot_to_png()
		self.compare(filename, get_comparison_image("empty1d"))

	def _assert_zoom_image(self):
		filename = self.window.plot_to_png()
		self.compare(filename, get_comparison_image("zoom1d"))

	def open_window(self):
		button = self.app.dataset_panel.button_histogram
		self.assert_(len(self.app.windows) == 0)
		QtTest.QTest.mouseClick(button, QtCore.Qt.LeftButton)

	def test_x(self):
		#QtTest.QTest.qWait(self.window.queue_update.default_delay)
		self.window._wait()
		filename = self.window.plot_to_png()
		self.compare(filename, get_comparison_image("example_x"))

	def test_r(self):
		self.layer.x = "sqrt(x**2+y**2)"
		self.window._wait()
		filename = self.window.plot_to_png()
		self.compare(filename, get_comparison_image("example_r"))

class TestPlotPanel2d(NoTest.TestPlotPanel):
	"""
	:type window: PlotDialog
	"""
	def open_window(self):
		button = self.app.dataset_panel.button_2d
		self.assert_(len(self.app.windows) == 0)
		QtTest.QTest.mouseClick(button, QtCore.Qt.LeftButton)

	def _move_to_empty_region(self):
		y = x = 1e4
		self.window.current_layer.xlim = [-10+x, 10+x]
		self.window.current_layer.ylim = [-10+y, 10+y]

	def _do_zoom(self):
		self.window.current_layer.xlim = [-10, 10]
		self.window.current_layer.ylim = [-10, 10]

	def _assert_default_image(self):
		filename = self.window.plot_to_png()
		self.compare(filename, get_comparison_image("default2d"))

	def _assert_empty_image(self):
		filename = self.window.plot_to_png()
		self.compare(filename, get_comparison_image("empty2d"))

	def _assert_zoom_image(self):
		filename = self.window.plot_to_png()
		self.compare(filename, get_comparison_image("zoom2d"))


	def test_xy(self):
		#QtTest.QTest.qWait(self.window.queue_update.default_delay)
		self.window._wait()
		filename = self.window.plot_to_png()
		self.compare(filename, get_comparison_image("example_xy"))

	def test_xr(self):
		self.layer.y = "sqrt(x**2+y**2)"
		self.window._wait()
		filename = self.window.plot_to_png()
		self.compare(filename, get_comparison_image("example_xr"))

	def test_xy_weight_r(self):
		self.layer.weight = "sqrt(x**2+y**2)"
		self.layer.statistic = "mean"
		self.layer.amplitude = "clip(grid, 0, 40)"
		self.window._wait()
		filename = self.window.plot_to_png()
		self.compare(filename, get_comparison_image("example_xy_weight_r"))

	def test_xy_vxvy(self):
		self.layer.vx = "vx"
		self.layer.vy = "vy"
		self.window._wait()
		filename = self.window.plot_to_png()
		self.compare(filename, get_comparison_image("example_xy_vxvy"))

		counter = self.window.queue_update.counter
		# the following actions should not cause an update
		self.layer.vx = "vx"
		self.layer.vy = "vy"
		self.assertEqual(counter, self.window.queue_update.counter)
		self.layer.vx = ""
		self.assertEqual(counter, self.window.queue_update.counter)
		self.layer.vx = None
		self.assertEqual(counter, self.window.queue_update.counter)
		self.layer.vx = ""
		self.assertEqual(counter, self.window.queue_update.counter)
		# this should update it
		self.layer.vx = "vx"
		self.assertEqual(counter+1, self.window.queue_update.counter)
		self.window._wait()

	def test_xy_vxvy_as_option(self):
		self.window.remove_layer()
		self.window.add_layer(["x", "y"], vx="vx", vy="vy", colorbar="False")
		self.window._wait()
		filename = self.window.plot_to_png()
		self.compare(filename, get_comparison_image("example_xy_vxvy"))

	def test_select_by_expression(self):
		self.window.xlabel = "x"
		self.window.ylabel = "y"
		##self.window._wait() # TODO: is this a bug? if we don't wait and directly do the selection, the ThreadPoolIndex
		## is entered twice, not sure this can happen from the gui
		expression = "x < 0"
		vaex.ui.qt.set_choose(expression, True)
		logger.debug("click mouse")
		QtTest.QTest.mouseClick(self.layer.button_selection_expression, QtCore.Qt.LeftButton)
		logger.debug("clicked mouse")
		return
		self.window._wait()
		self.assertTrue(self.no_exceptions)

		filename = self.window.plot_to_png()
		self.compare(filename, get_comparison_image("example_xy_selection_on_x"))

	def test_selection_options(self):
		self.window.xlabel = "x"
		self.window.ylabel = "y"
		##self.window._wait() # TODO: is this a bug? if we don't wait and directly do the selection, the ThreadPoolIndex
		## is entered twice, not sure this can happen from the gui
		expression = "x < 0"
		vaex.ui.qt.set_choose(expression, True)
		QtTest.QTest.mouseClick(self.layer.button_selection_expression, QtCore.Qt.LeftButton)

		# test if the expressions ends up in the clipboard
		self.window.action_selection_copy.trigger()
		clipboard = QtGui.QApplication.clipboard()
		clipboard_text = clipboard.text()
		self.assertIn(expression, clipboard_text)

		self.window._wait()
		self.assertTrue(self.no_exceptions)
		filename = self.window.plot_to_png()
		self.compare(filename, get_comparison_image("example_xy_selection_on_x"))


		# select nothing
		self.window.action_select_none.trigger()
		self.window._wait()
		filename = self.window.plot_to_png()
		self.compare(filename, get_comparison_image("example_xy"))
		with dialogs.assertError(1):
			self.window.action_selection_copy.trigger()

		# paste the expression back
		print("paste back the expression")
		clipboard.setText(clipboard_text)
		self.window.action_selection_paste.trigger()
		print("waiting....")
		self.window._wait()
		filename = self.window.plot_to_png()
		self.compare(filename, get_comparison_image("example_xy_selection_on_x"))



		#clipboard.setText("junk")
		#with dialogs.assertError(1):
		#	self.window.action_selection_paste.trigger()
			#self.window.action_selection_copy.trigger()




	def test_select_by_lasso(self):
		self.window._wait() # TODO: is this a bug? same as above
		vaex.ui.qt.set_choose("x < 0", True)
		x = [-10, 10, 10, -10]
		y = [-10, -10, 10, 10]
		self.layer.dataset.select_lasso("x", "y", x, y)
		#QtTest.QTest.mouseClick(self.layer.button_selection_expression, QtCore.Qt.LeftButton)
		self.assertLess(self.layer.dataset.selected_length(), len(self.layer.dataset))
		self.window._wait()
		self.assertTrue(self.no_exceptions)

	def test_layers(self):
		self.window.add_layer(["x", "z"])
		self.window._wait()

	def test_resolution(self):
		if 0: # keyClick doesn't work on osx it seems
			self.window.show()
			QtTest.QTest.qWaitForWindowShown(self.window)
			QtTest.QTest.keyClick(self.window, QtCore.Qt.Key_1, QtCore.Qt.ControlModifier|QtCore.Qt.AltModifier) # "Ctrl+Alt+1"should be 32x32
		else:
			self.window.action_resolution_list[0].trigger()
		self.window._wait()
		filename = self.window.plot_to_png()
		self.compare(filename, get_comparison_image("example_xy_32x32"))

	def test_resolution_vector(self):
		self.layer.vx = "vx"
		self.layer.vy = "vy"
		self.window.action_resolution_vector_list[2].trigger()
		self.window._wait()
		filename = self.window.plot_to_png()
		self.compare(filename, get_comparison_image("example_xy_vxvy_32x32"))


	def test_invalid_expression(self):
		self.window._wait()

		with dialogs.assertError(2):
			self.layer.x = "vx*"
			self.layer.y = "vy&"
		with dialogs.assertError(3):
			self.layer.x = "hoeba(vx)"
			self.layer.x = "x(vx)"
			self.layer.y = "doesnotexist"
		with dialogs.assertError(2):
			self.layer.vx = "hoeba(vx)"
			self.layer.vy = "x(vx)"
		with dialogs.assertError(1):
			self.layer.weight = "hoeba(vx)"
			self.layer.statistic = "mean"
		with dialogs.assertError(1):
			self.layer.statistic = "blaat"
		self.layer.x = "x"
		self.layer.y = "y"
		self.layer.weight = "z"
		self.layer.statistic = "mean"
		#self.window._wait()
		# since this will be triggered, overrule it
		self.no_error_in_field = True

import sys
test_port = 29310 + sys.version_info[0] * 10 + sys.version_info[1]

class TestPlotPanel2dRemote(TestPlotPanel2d):
	use_websocket = False
	@classmethod
	def setUpClass(cls):
		global test_port
		cls.webserver = vaex.webserver.WebServer(datasets=[], port=test_port, cache_byte_size=0)
		#print "serving"
		cls.webserver.serve_threaded()
		#print "getting server object"
		scheme = "ws" if cls.use_websocket else "http"
		cls.server = vx.server("%s://localhost:%d" % (scheme, test_port))
		test_port += 1


	@classmethod
	def tearDownClass(cls):
		cls.server.close()
		cls.webserver.stop_serving()

	def create_app(self):
		logger.debug("create app")
		global test_port
		self.app = vx.ui.main.VaexApp([], open_default=False)
		self.dataset_default = vaex.example()
		datasets = [self.dataset_default]
		self.webserver.set_datasets(datasets)
		#self.webserver = vaex.webserver.WebServer(datasets=datasets, port=test_port)
		#print "serving"
		logger.debug("serve server")
		#self.webserver.serve_threaded()
		#print "getting server object"
		#scheme = "ws" if self.use_websocket else "http"
		#logger.debug("get from server")
		#self.server = vx.server("%s://localhost:%d" % (scheme, test_port), thread_mover=self.app.call_in_main_thread)
		self.server.thread_mover=self.app.call_in_main_thread
		datasets = self.server.datasets(as_dict=True)
		logger.debug("got it")

		self.dataset = datasets[self.dataset_default.name]
		self.app.dataset_selector.add(self.dataset)
		test_port += 1
		logger.debug("create app done")

	def tearDown(self):
		logger.debug("closing all")
		#print "stop serving"
		TestPlotPanel2d.tearDown(self)
		#self.webserver.stop_serving()
		#self.server.close()


	def test_select_by_lasso(self):
		pass # TODO: cannot test since DatasetRemote.selected_length it not implemented

	#def test_invalid_expression(self): pass
	#def test_resolution_vector(self): pass
	#def test_resolution(self): pass
	#def test_layers(self): pass
	#def test_select_by_lasso(self): pass
	#def test_select_by_expression(self): pass
	#def test_xy_vxvy_as_option(self): pass
	#def test_xy_vxvy(self): pass
	#def test_xy_weight_r(self): pass
	#def test_xr(self): pass
	#def test_xy(self): pass

if __name__ == '__main__':
    unittest.main()



import functools
import vaex.vaexfast
import numpy as np
from vaex.utils import Timer
import vaex.events
from . import multithreading
import time
import math
#import vaex.ui.expressions as expr
from functools import reduce

__author__ = 'breddels'

buffer_size_default = 1024*1024 # TODO: this should not be fixed, larger means faster but also large memory usage
#buffer_size_default = 1e4

import threading
import queue
import math
import multiprocessing
import sys
import collections
lock = threading.Lock()
import vaex.multithreading

import logging
logger = logging.getLogger("vaex.execution")



thread_count_default = multiprocessing.cpu_count()

class UserAbort(Exception):
	def __init__(self, reason):
		super(UserAbort, self).__init__(reason)
class Column(collections.namedtuple('Column', ['dataset', 'expression'])):
	def needs_copy(self):
		return not \
			(self.expression in self.dataset.column_names  \
			and not isinstance(self.dataset.columns[self.expression], vaex.dataset._ColumnConcatenatedLazy)\
			and self.dataset.columns[self.expression].dtype.type==np.float64 \
			and self.dataset.columns[self.expression].strides[0] == 8 \
			and self.expression not in self.dataset.virtual_columns)
				#and False:

class Job(object):
	def __init__(self, task, order):
		self.task = task
		self.order = order

# mutex for numexpr, it is not thread save
ne_lock = threading.Lock()

class Executor(object):
	def __init__(self, thread_pool=None, buffer_size=None, thread_mover=None):
		self.thread_pool = thread_pool or vaex.multithreading.ThreadPoolIndex()
		self.task_queue = []
		self.buffer_size = buffer_size or buffer_size_default
		self.signal_begin = vaex.events.Signal("begin")
		self.signal_progress = vaex.events.Signal("progress")
		self.signal_end = vaex.events.Signal("end")
		self.signal_cancel = vaex.events.Signal("cancel")
		self.thread_mover = thread_mover or (lambda fn, *args, **kwargs: fn(*args, **kwargs))
		self._is_executing = False
		self.lock = threading.Lock()
		self.thread = None
		self.passes = 0  # how many times we passed over the data

	def schedule(self, task):
		self.task_queue.append(task)
		logger.info("task added, queue: %r", self.task_queue)
		return task

	def run(self, task):
		#with self.lock:
		if 1:
			logger.debug("added task: %r" % task)
			previous_queue = self.task_queue
			try:
				self.task_queue = [task]
				self.execute()
			finally:
				self.task_queue = previous_queue
			return task._value

	def execute_threaded(self):
		if self.thread is None:
			logger.info("starting thread for executor")
			self.thread = threading.Thread(target=self._execute_in_thread)
			self.thread.start()
			self.queue_semaphore = threading.Semaphore()
		logger.info("sending thread a msg that it can execute")
		self.queue_semaphore.release()

	def _execute_in_thread(self):
		while True:
			try:
				logger.info("waiting for jobs")
				self.queue_semaphore.acquire()
				logger.info("got jobs")
				if self.task_queue:
					logger.info("executing tasks in thread: %r", self.task_queue)
					self.execute()
				else:
					logger.info("empty task queue")
			except:
				import traceback
				traceback.print_exc()
				logger.error("exception occured in thread")

	def execute(self):
		logger.debug("starting with execute")
		if self._is_executing:
			logger.debug("nested execute call")
			# this situation may happen since in this methods, via a callback (to update a progressbar) we enter
			# Qt's eventloop, which may execute code that will call execute again
			# as long as that code is using delay tasks (i.e. promises) we can simple return here, since after
			# the execute is almost finished, any new tasks added to the task_queue will get executing
			return
		# u 'column' is uniquely identified by a tuple of (dataset, expression)
		self._is_executing = True
		try:
			t0 = time.time()
			task_queue_all = list(self.task_queue)
			if not task_queue_all:
				logger.info("only had cancelled tasks")
			logger.info("clearing queue")
			#self.task_queue = [] # Ok, this was stupid.. in the meantime there may have been new tasks, instead, remove the ones we copied
			for task in task_queue_all:
				logger.info("remove from queue: %r", task)
				self.task_queue.remove(task)
			logger.info("left in queue: %r", self.task_queue)
			task_queue_all = [task for task in task_queue_all if not task.cancelled]
			logger.debug("executing queue: %r" % (task_queue_all))

			#for task in self.task_queue:
			#$	print task, task.expressions_all
			datasets = set(task.dataset for task in task_queue_all)
			cancelled = [False]
			def cancel():
				logger.debug("cancelling")
				self.signal_cancel.emit()
				cancelled[0] = True
			try:
				# process tasks per dataset
				self.signal_begin.emit()
				for dataset in datasets:
					self.passes += 1
					task_queue = [task for task in task_queue_all if task.dataset == dataset]
					expressions = list(set(expression for task in task_queue for expression in task.expressions_all))

					for task in task_queue:
						task._results = []
						task.signal_progress.emit(0)
					block_scopes = [dataset._block_scope(0, self.buffer_size) for i in range(self.thread_pool.nthreads)]
					def process(thread_index, i1, i2):
						if not cancelled[0]:
							block_scope = block_scopes[thread_index]
							block_scope.move(i1 + dataset._index_start, i2 + dataset._index_start)
							#with ne_lock:
							block_dict = {expression:block_scope.evaluate(expression) for expression in expressions}
							for task in task_queue:
								blocks = [block_dict[expression] for expression in task.expressions_all]
								if not cancelled[0]:
									task._results.append(task.map(thread_index, i1, i2, *blocks))
								# don't call directly, since ui's don't like being updated from a different thread
								#self.thread_mover(task.signal_progress, float(i2)/length)
#								time.sleep(0.1)

					length = dataset.active_length()
					#print self.thread_pool.map()
					for element in self.thread_pool.map(process, vaex.utils.subdivide(length, max_length=self.buffer_size),\
														progress=lambda p: all(self.signal_progress.emit(p)) and\
																all([all(task.signal_progress.emit(p)) for task in task_queue]),
														cancel=cancel, unpack=True):
						pass # just eat all element
					self._is_executing = False
			except:
				# on any error we flush the task queue
				self.signal_cancel.emit()
				logger.exception("error in task, flush task queue")
				raise
			logger.debug("executing took %r seconds" % (time.time() - t0))
			# while processing the self.task_queue, new elements will be added to it, so copy it
			logger.debug("cancelled: %r", cancelled)
			if cancelled[0]:
				logger.debug("execution aborted")
				task_queue = task_queue_all
				for task in task_queue:
					#task._result = task.reduce(task._results)
					#task.reject(UserAbort("cancelled"))
					# remove references
					task._result = None
					task._results = None
			else:
				task_queue = task_queue_all
				for task in task_queue:
					logger.debug("fulfill task: %r", task)
					if not task.cancelled:
						task._result = task.reduce(task._results)
						task.fulfill(task._result)
						# remove references
					task._result = None
					task._results = None
				self.signal_end.emit()
				# if new tasks were added as a result of this, execute them immediately
				# TODO: we may want to include infinite recursion protection
				self._is_executing = False
				if len(self.task_queue) > 0:
					logger.debug("task queue not empty.. start over!")
					self.execute()
		finally:
			self._is_executing = False

#default_executor = None



__author__ = 'maartenbreddels'
import os
import mmap
import itertools
import functools
import collections
import logging
import numpy as np
import numpy.ma
import vaex
import astropy.table
import astropy.units
from vaex.utils import ensure_string
import astropy.io.fits as fits
import re
import six

from vaex.dataset import DatasetLocal, DatasetArrays
logger = logging.getLogger("vaex.file")
import vaex.dataset
import vaex.file
dataset_type_map = {}

# h5py doesn't want to build at readthedocs
on_rtd = os.environ.get('READTHEDOCS', None) == 'True'
try:
	import h5py
except:
	if not on_rtd:
		raise

from vaex.expression import Expression

class DatasetMemoryMapped(DatasetLocal):
	"""Represents a dataset where the data is memory mapped for efficient reading"""


	# nommap is a hack to get in memory datasets working
	def __init__(self, filename, write=False, nommap=False, name=None):
		super(DatasetMemoryMapped, self).__init__(name=name or os.path.splitext(os.path.basename(filename))[0], path=os.path.abspath(filename) if filename is not None else None, column_names=[])
		self.filename = filename or "no file"
		self.write = write
		#self.name = name or os.path.splitext(os.path.basename(self.filename))[0]
		#self.path = os.path.abspath(filename) if filename is not None else None
		self.nommap = nommap
		if not nommap:
			self.file = open(self.filename, "r+" if write else "r")
			self.fileno = self.file.fileno()
			kwargs = {}
			if vaex.utils.osname == "windows":
				kwargs["access"]=mmap.ACCESS_READ | 0 if not write else mmap.ACCESS_WRITE
			else:
				kwargs["prot"]=mmap.PROT_READ | 0 if not write else mmap.PROT_WRITE
			self.mapping = mmap.mmap(self.fileno, 0, **kwargs)
			self.file_map = {filename: self.file}
			self.fileno_map = {filename: self.fileno}
			self.mapping_map = {filename: self.mapping}
		else:
			self.file_map = {}
			self.fileno_map = {}
			self.mapping_map = {}
		self._length_original = None
		#self._fraction_length = None
		self.nColumns = 0
		self.column_names = []
		self.rank1s = {}
		self.rank1names = []
		self.virtual_columns = collections.OrderedDict()

		self.axes = {}
		self.axis_names = []

		# these are replaced by variables
		#self.properties = {}
		#self.property_names = []

		self.current_slice = None
		self.fraction = 1.0


		self.selected_row_index = None
		self.selected_serie_index = 0
		self.row_selection_listeners = []
		self.serie_index_selection_listeners = []
		#self.mask_listeners = []

		self.all_columns = {}
		self.all_column_names = []
		self.global_links = {}

		self.offsets = {}
		self.strides = {}
		self.filenames = {}
		self.dtypes = {}
		self.samp_id = None
		#self.variables = collections.OrderedDict()

	def close_files(self):
		for name, file in self.file_map.items():
			file.close()
		# on osx and linux this will give random bus errors (osx) or segfaults (linux)
		# on win32 however, we'll run out of file handles
		if vaex.utils.osname not in ["osx", "linux"]:
			for name, memmap in self.mapping_map.items():
				memmap.close()

	def has_snapshots(self):
		return len(self.rank1s) > 0

	def get_path(self):
		return self.path

	def addFile(self, filename, write=False):
		self.file_map[filename] = open(filename, "r+" if write else "r")
		self.fileno_map[filename] = self.file_map[filename].fileno()
		self.mapping_map[filename] = mmap.mmap(self.fileno_map[filename], 0, prot=mmap.PROT_READ | 0 if not write else mmap.PROT_WRITE )


	def selectSerieIndex(self, serie_index):
		self.selected_serie_index = serie_index
		for serie_index_selection_listener in self.serie_index_selection_listeners:
			serie_index_selection_listener(serie_index)
		self.signal_sequence_index_change.emit(self, serie_index)

	def matches_url(self, url):
		filename = url
		if filename.startswith("file:/"):
			filename = filename[5:]
		similar = os.path.splitext(os.path.abspath(self.filename))[0] == os.path.splitext(filename)[0]
		logger.info("matching urls: %r == %r == %r" % (os.path.splitext(self.filename)[0], os.path.splitext(filename)[0], similar) )
		return similar

	def close(self):
		self.file.close()
		self.mapping.close()

	def addAxis(self, name, offset=None, length=None, dtype=np.float64, stride=1, filename=None):
		if filename is None:
			filename = self.filename
		mapping = self.mapping_map[filename]
		mmapped_array = np.frombuffer(mapping, dtype=dtype, count=length if stride is None else length * stride, offset=offset)
		if stride:
			mmapped_array = mmapped_array[::stride]
		self.axes[name] = mmapped_array
		self.axis_names.append(name)


	def addColumn(self, name, offset=None, length=None, dtype=np.float64, stride=1, filename=None, array=None):
		if filename is None:
			filename = self.filename
		if not self.nommap:
			mapping = self.mapping_map[filename]

		if array is not None:
			length = len(array)

		if self._length_original is not None and length != self._length_original:
			logger.error("inconsistent length", "length of column %s is %d, while %d was expected" % (name, length, self._length))
		else:
			self._length_unfiltered = length
			self._length_original = length
			if self.current_slice is None:
				self.current_slice = (0, length)
				self.fraction = 1.
				self._length = length
				self._index_end = self._length_unfiltered
				self._index_start = 0
			#print self.mapping, dtype, length if stride is None else length * stride, offset
			if array is not None:
				length = len(array)
				mmapped_array = array
				stride = None
				offset = None
				dtype = array.dtype
			else:
				if offset is None:
					print("offset is None")
					sys.exit(0)
				mmapped_array = np.frombuffer(mapping, dtype=dtype, count=length if stride is None else length * stride, offset=offset)
				if stride:
					#import pdb
					#pdb.set_trace()
					mmapped_array = mmapped_array[::stride]
			self.columns[name] = mmapped_array
			self.column_names.append(name)
			self._save_assign_expression(name, Expression(self, name))
			self.all_columns[name] = mmapped_array
			self.all_column_names.append(name)
			#self.column_names.sort()
			self.nColumns += 1
			self.nRows = self._length_original
			self.offsets[name] = offset
			self.strides[name] = stride
			if filename is not None:
				self.filenames[name] = os.path.abspath(filename)
			self.dtypes[name] = dtype

	def addRank1(self, name, offset, length, length1, dtype=np.float64, stride=1, stride1=1, filename=None, transposed=False):
		if filename is None:
			filename = self.filename
		mapping = self.mapping_map[filename]
		if (not transposed and self._length is not None and length != self._length) or (transposed and self._length is not None and length1 != self._length):
			logger.error("inconsistent length", "length of column %s is %d, while %d was expected" % (name, length, self._length))
		else:
			if self.current_slice is None:
				self.current_slice = (0, length if not transposed else length1)
				self.fraction = 1.
				self._length_unfiltered = length if not transposed else length1
				self._length = self._length_unfiltered
				self._index_end = self._length_unfiltered
			self._length = length if not transposed else length1
			#print self.mapping, dtype, length if stride is None else length * stride, offset
			rawlength = length * length1
			rawlength *= stride
			rawlength *= stride1

			mmapped_array = np.frombuffer(mapping, dtype=dtype, count=rawlength, offset=offset)
			mmapped_array = mmapped_array.reshape((length1*stride1, length*stride))
			mmapped_array = mmapped_array[::stride1,::stride]

			self.rank1s[name] = mmapped_array
			self.rank1names.append(name)
			self.all_columns[name] = mmapped_array
			self.all_column_names.append(name)



import struct
class HansMemoryMapped(DatasetMemoryMapped):
	def __init__(self, filename, filename_extra=None):
		super(HansMemoryMapped, self).__init__(filename)
		self.pageSize, \
		self.formatSize, \
		self.numberParticles, \
		self.numberTimes, \
		self.numberParameters, \
		self.numberCompute, \
		self.dataOffset, \
		self.dataHeaderSize = struct.unpack("Q"*8, self.mapping[:8*8])
		zerooffset = offset = self.dataOffset
		length = self.numberParticles+1
		stride = self.formatSize//8 # stride in units of the size of the element (float64)

		# TODO: ask Hans for the self.numberTimes-2
		lastoffset = offset + (self.numberParticles+1)*(self.numberTimes-2)*self.formatSize
		t_index = 3
		names = "x y z vx vy vz".split()
		midoffset = offset + (self.numberParticles+1)*self.formatSize*t_index
		names = "x y z vx vy vz".split()

		for i, name in enumerate(names):
			self.addColumn(name+"_0", offset+8*i, length, dtype=np.float64, stride=stride)

		for i, name in enumerate(names):
			self.addColumn(name+"_last", lastoffset+8*i, length, dtype=np.float64, stride=stride)


		names = "x y z vx vy vz".split()

		if 1:
			stride = self.formatSize//8
			#stride1 = self.numberTimes #*self.formatSize/8
			for i, name in enumerate(names):
				# TODO: ask Hans for the self.numberTimes-1
				self.addRank1(name, offset+8*i, length=self.numberParticles+1, length1=self.numberTimes-1, dtype=np.float64, stride=stride, stride1=1)

		if filename_extra is None:
			basename = os.path.basename(filename)
			if os.path.exists(basename + ".omega2"):
				filename_extra = basename + ".omega2"

		if filename_extra is not None:
			self.addFile(filename_extra)
			mapping = self.mapping_map[filename_extra]
			names = "J_r J_theta J_phi Theta_r Theta_theta Theta_phi Omega_r Omega_theta Omega_phi r_apo r_peri".split()
			offset = 0
			stride = 11
			#import pdb
			#pdb.set_trace()
			for i, name in enumerate(names):
				# TODO: ask Hans for the self.numberTimes-1
				self.addRank1(name, offset+8*i, length=self.numberParticles+1, length1=self.numberTimes-1, dtype=np.float64, stride=stride, stride1=1, filename=filename_extra)

				self.addColumn(name+"_0", offset+8*i, length, dtype=np.float64, stride=stride, filename=filename_extra)
				self.addColumn(name+"_last", offset+8*i + (self.numberParticles+1)*(self.numberTimes-2)*11*8, length, dtype=np.float64, stride=stride, filename=filename_extra)


	@classmethod
	def can_open(cls, path, *args, **kwargs):
		return os.path.splitext(path)[-1] == ".bin"
		basename, ext = os.path.splitext(path)
		#if os.path.exists(basename + ".omega2"):
		#	return True
		#return True

	@classmethod
	def get_options(cls, path):
		return []

	@classmethod
	def option_to_args(cls, option):
		return []
dataset_type_map["buist"] = HansMemoryMapped


__author__ = 'maartenbreddels'
import aplus

"""
Mini library to make working with promises/futures a bit more Python like.

Example:

@delayed
def f(grid):
	return grid**2

f(ds.count(delay=True))

"""

def promisify(value):
	# TODO, support futures etc..
	if isinstance(value, aplus.Promise):
		return value
	if isinstance(value, (list, tuple)):
		return aplus.listPromise(*list([promisify(k) for k in value]))
	else:
		return aplus.Promise.fulfilled(value)
def delayed(f):
	def wrapped(*args, **kwargs):
		#print "calling", f, "with", kwargs
		#key_values = kwargs.items()
		key_promise = list([(key, promisify(value)) for key, value in kwargs.items()])
		#key_promise = [(key, promisify(value)) for key, value in key_values]
		arg_promises = list([promisify(value) for value in args])
		kwarg_promises = list([promise for key, promise in key_promise])
		promises = arg_promises + kwarg_promises
		for promise in promises:
			def echo_error(exc, promise=promise):
				print("error with ", promise, "exception is", exc)
				#raise exc
			def echo(value, promise=promise):
				print("done with ", repr(promise), "value is", value)
			#promise.then(echo, echo_error)

		#print promises
		allarguments = aplus.listPromise(*promises)
		def call(_):
			kwargs_real = {key:promise.get() for key, promise in key_promise}
			args_real = list([promise.get() for promise in arg_promises])
			return f(*args_real, **kwargs_real)
		def error(exc):
			print("error", exc)
			raise exc
		return allarguments.then(call, error)
	return wrapped

@delayed
def delayed_args(*args):
	return args

@delayed
def delayed_list(l):
	return delayed_args(*l)


from __future__ import absolute_import
__author__ = 'maartenbreddels'

import tornado.ioloop
import tornado.web
import tornado.httpserver
import tornado.websocket
import tornado.auth
import tornado.gen
import threading
import logging
import vaex as vx
import vaex.utils
import json
import inspect
import yaml
import argparse
import os
import time
import numpy as np

logger = logging.getLogger("vaex.webserver")
job_index = 0
class JobFlexible(object):
	def __init__(self, cost, fn=None, args=None, kwargs=None, index=None):
		global job_index
		self.cost = cost
		if index is None:
			self.index = job_index
			job_index += 1
		else:
			self.index = index
		self.fraction = None
		self.time_elapsed_goal = None
		self.time_created = time.time()
		self.fn = fn or (lambda *args, **kwargs: None)
		self.args = args
		self.kwargs = kwargs

	def set_fraction(self, fraction):
		self.fraction = fraction

	def set_time_elapsed_goal(self, time_elapsed_goal):
		self.time_elapsed_goal = time_elapsed_goal

	def execute(self):
		try:
			result = self.fn(*self.args, **self.kwargs)
			return result
		finally:
			self.done()

	def done(self):
		self.time_done = time.time()
		self.time_elapsed = self.time_done - self.time_created
import vaex.events
class JobQueue(object):
	def __init__(self):
		self.queue = []
		self.queue_delay = 1. # 1 second to get rid of the queue
		self.seconds_per_cost = 1.
		self.cost_modifier_per_miss = 0.1
		self.fraction_minimum = 1e-4
		self.signal_job_finished = vaex.events.Signal("job finished")

	def add(self, job):
		self.queue.append(job)

	def get_next(self):
		if len(self.queue) == 0:
			raise IndexError("queue empty")
		total_cost = sum(job.cost for job in self.queue)
		job_count = len(self.queue)
		job = self.queue.pop(0)
		logger.debug("queue length is %d, total cost of queue is %f, job index %d", len(self.queue), total_cost, job.index)
		# if the total cost equals seconds, and we want the queue to be finished in 1 second
		# this will be the best fraction
		elapsed = time.time() - job.time_created
		if elapsed > self.queue_delay * 1.3: # we already spent way too long time, the math doesn't work out, so what now?
			# we scale the cost up, since it apearently doesn't work out fully
			#raise ValueError, "elapsed = %f" % elapsed
			logger.error("queue clogging up")
		# assuming the queue costs will stay constant we want to finished this whole queue in max self.queue_delay
		# seconds. Therefore the current job should only calculate a fraction as computed by:
		fraction = (self.queue_delay - elapsed) / (total_cost * self.seconds_per_cost) * 0.9
		fraction = min(max(self.fraction_minimum, fraction), 1)
		time_elapsed_goal = elapsed + job.cost * self.seconds_per_cost * fraction / 0.9#elapsed + fraction * (self.seconds_per_cost)
		logger.debug("to finish the queue we set the next job's fraction to %f (time elapsed = %f)", fraction, elapsed)
		#print fraction
		job.set_fraction(fraction)
		job.set_time_elapsed_goal(time_elapsed_goal)
		return job

	def finished(self, job):
		ratio = job.time_elapsed/ job.time_elapsed_goal # / job.time_elapsed_goal
		if ratio > 1.2:
			logger.debug("job should have taken %f seconds, took %f seconds, we should go faster", job.time_elapsed_goal, job.time_elapsed)
			self.seconds_per_cost *= (1 + self.cost_modifier_per_miss)
		if ratio < 0.75:
			self.seconds_per_cost *= (1 - self.cost_modifier_per_miss)
			logger.debug("job should have taken %f seconds, took %f seconds, we can slow down", job.time_elapsed_goal, job.time_elapsed)
		logger.debug("seconds_per_cost set to: %f", self.seconds_per_cost)
		self.signal_job_finished.emit(job)

		#seconds_goal = job.set_fraction
		#fraction_goal = job.set_fraction
		#fraction_achieved = job.ti
		pass

#import concurrent.future

class JobExecutor(object):
	def __init__(self, job_queue):
		self.job_queue = job_queue
		#self.thread_pool_executor = concurrent.

	def empty_queue(self):
		while True:
			try:
				job = self.job_queue.get_next()
				job.execute()
				self.job_queue.finished(job)
			except IndexError:
				return

class GoogleOAuth2LoginHandler(tornado.web.RequestHandler,
                               tornado.auth.GoogleOAuth2Mixin):
    @tornado.gen.coroutine
    def get(self):
        if self.get_argument('code', False):
            user = yield self.get_authenticated_user(
                redirect_uri='http://your.site.com/auth/google',
                code=self.get_argument('code'))
            # Save the user with e.g. set_secure_cookie
        else:
            yield self.authorize_redirect(
                redirect_uri='http://your.site.com/auth/google',
                client_id='cid',
                scope=['profile', 'email'],
                response_type='code',
                extra_params={'approval_prompt': 'auto'})

def task_invoke(subspace, method_name, **arguments):
	method = getattr(subspace, method_name)
	#args, varargs, kwargs, defaults = inspect.getargspec(method)
	#print inspect.getargspec(method)
	#args_required = args[:len(args)-len(defaults)]
	#kwargs = {}
	#for arg in args:
	#	if arg in arguments:
	#		kwargs[arg] = arguments[arg]
	#print("!!!!!! calling %s with arguments %r" % (method_name, arguments))
	values = method(**arguments)
	return values

import collections
import concurrent.futures
import vaex.execution
import vaex.multithreading
import tornado.escape
class ListHandler(tornado.web.RequestHandler):
	def initialize(self, webserver, submit_threaded, cache, cache_selection, datasets=None):
		self.webserver = webserver
		self.submit_threaded = submit_threaded
		self.cache = cache
		self.cache_selection = cache_selection

	def set_default_headers(self):
		self.set_header("Access-Control-Allow-Origin", "*")

	@tornado.gen.coroutine
	def get(self):
		user_id = None
		if False: # disable user_id for the moment
			user_id = self.get_cookie("user_id")
			logger.debug("user_id in cookie: %r", user_id)
			if user_id is None:
				import uuid
				user_id = str(uuid.uuid4())
				self.set_cookie("user_id", user_id)
			logger.debug("user_id: %r", user_id)
		# tornado retursn a list of values, just use the first value
		arguments = {key:json.loads(tornado.escape.to_basestring(self.request.arguments[key][0])) for key in self.request.arguments.keys()}
		key = (self.request.path, "-".join([str(v) for v in sorted(arguments.items())]))
		response = self.cache.get(key)
		if response is None:
			response = yield self.submit_threaded(process, self.webserver, user_id, self.request.path, **arguments)
			try:
				self.cache[key] = response
				pass
			except ValueError:
				pass # raised when it doesn't fit in cache
		else:
			logger.debug("cache hit for key: %r", key)
		#logger.debug("response is: %r", response)
		if response is None:
			response = self.error("unknown request or error")
		if isinstance(response, (np.ndarray, float)):
			response = np.array(response)
			self.set_header("Content-Type", "application/numpy-array")
			self.write(str(response.shape)+"\n")
			self.write(str(response.dtype)+"\n")
			self.write(response.tobytes())
		else:
			if isinstance(response, str):
				self.set_header("Content-Type", "application/octet-stream")
			self.write(response)


class ProgressWebSocket(tornado.websocket.WebSocketHandler):
	def initialize(self, webserver, submit_threaded, cache, cache_selection, datasets=None):
		self.webserver = webserver
		self.submit_threaded = submit_threaded
		self.cache = cache
		self.cache_selection = cache_selection

	def check_origin(self, origin):
		return True

	def open(self):
		logger.debug("WebSocket opened")

	@tornado.gen.coroutine
	def on_message(self, message):
		logger.debug("websocket message: %r", message)
		arguments = json.loads(message)
		path = arguments["path"]
		arguments.pop("path")
		user_id = arguments.pop("user_id", None)
		job_id = arguments.pop("job_id")


		if False: # disable user_id for the moment
			user_id = self.get_cookie("user_id")
			if user_id == None:
				user_id = arguments.pop("user_id")
				if user_id == None:
					self.write_json(error="no user id, do an http get request first")
					return
		last_progress = [None]
		def progress(f):
			def do():
				logger.debug("progress: %r", f)
				last_progress[0] = f
				self.write_json(job_id=job_id, job_phase="PENDING", progress=f)
			if last_progress[0] is None or (f - last_progress[0]) > 0.05 or f == 1.0:
				tornado.ioloop.IOLoop.current().add_callback(do)
			return True
		key = (path, "-".join([str(v) for v in sorted(arguments.items())]))
		response = self.cache.get(key)
		progress(0)
		if response is None:
			#response = yield self.submit_threaded(process, self.webserver, user_id, self.request.path, **arguments)
			response = yield self.submit_threaded(process, self.webserver, user_id, path, progress=progress,
												  **arguments)
			try:
				self.cache[key] = response
				pass
			except ValueError:
				pass # raised when it doesn't fit in cache
		progress(1)

		if response is None:
			response = self.error("unknown request or error")

		meta_data = dict(job_id=job_id)
		if isinstance(response, (np.ndarray, float)):
			response = np.array(response)
			#self.set_header("Content-Type", "application/numpy-array")
			meta_data["shape"] = str(response.shape)
			meta_data["dtype"] = str(response.dtype)
			meta_data["job_phase"] = "COMPLETED"
			import zlib
			data_response = zlib.compress(response.tobytes())
		else: # it should be a dict
			if "result" in response:
				response["job_phase"] = "COMPLETED"
			if "error" in response:
				response["job_phase"] = "ERROR"
			if "exception" in response:
				response["job_phase"] = "EXCEPTION"
			meta_data.update(response)
			data_response = b""
		try:
			text_response = json.dumps(meta_data).encode("utf8")\
						+ b"\n" + data_response
		except Exception as e:
			meta_data = dict(job_id=job_id, job_phase='EXCEPTION', **exception(e))
			text_response = json.dumps(meta_data).encode("utf8")\
						+ b"\n" + data_response

		self.write_message(text_response, binary=True)

	def write_json(self, **kwargs):
		logger.debug("writing json: %r", kwargs)
		self.write_message(json.dumps(kwargs) + "\n", binary=True)

	def on_close(self):
		logger.debug("WebSocket closed")

class QueueHandler(tornado.web.RequestHandler):
	def initialize(self, datasets):
		self.datasets = datasets
		self.datasets_map = dict([(ds.name,ds) for ds in self.datasets])

	def get(self):
        #self.write("Hello, world")
		self.write(dict(datasets=[{"name":ds.name, "length":len(ds)} for ds in self.datasets]))

def exception(exception):
	logger.exception("handled exception at server, all fine")
	return ({"exception": {"class":str(exception.__class__.__name__), "msg": str(exception)} })

def error(msg):
	return ({"error": msg}) #, "result":None})

def process(webserver, user_id, path, fraction=None, progress=None, **arguments):
	if not hasattr(webserver.thread_local, "executor"):
		logger.debug("creating thread pool and executor")
		webserver.thread_local.thread_pool = vaex.multithreading.ThreadPoolIndex(max_workers=webserver.threads_per_job)
		webserver.thread_local.executor = vaex.execution.Executor(thread_pool=webserver.thread_local.thread_pool)
		webserver.thread_pools.append(webserver.thread_local.thread_pool)

	progress = progress or (lambda x: True)
	progress(0)
	# TODO: mem leak and other issues if we don't disconnect this
	webserver.thread_local.executor.signal_progress.connect(progress)
	#return ("Hello, world")
	#print request.path
	try:
		parts = [part for part in path.split("/") if part]
		logger.debug("request: %r" % parts)
		if parts[0] == "sleep":
			seconds = float(parts[1])
			import time
			time.sleep(min(1, seconds))
			return ({"result":seconds})
		elif parts[0] == "datasets":
			if len(parts) == 1:
				#
				response = dict(result=[{'name':ds.name,
										'length_original':ds.length_original(),
										'column_names':ds.get_column_names(strings=True),
										 'dtypes':{name:str(ds.columns[name].dtype) for name in ds.get_column_names(strings=True)},
										 'state': ds.state_get()
										 } for ds in webserver.datasets])
				logger.debug("response: %r", response)
				return response
			else:

				dataset_name = parts[1]
				logger.debug("dataset: %s", dataset_name)
				if dataset_name not in webserver.datasets_map:
					error("dataset does not exist: %r, possible options: %r" % (dataset_name, webserver.datasets_map.keys()))
				else:
					if len(parts) > 2:
						method_name = parts[2]
						if "expressions" in arguments:
							expressions = arguments["expressions"]
						else:
							expressions = None
						# make a shallow copy, such that selection and active_fraction is not shared
						dataset_original = webserver.datasets_map[dataset_name]
						dataset = dataset_original.copy()
						# copy the cache, 2 level deep clone
						# currently not working, could be a good optimization
						# dataset._selection_mask_caches = {key:dict(value) for key, value in dataset_original._selection_mask_caches.items()}
						logger.info("method: %r args: %r" % (method_name, arguments))
						# TODO: executor should be an argument to the stats functions..
						dataset.executor = webserver.thread_local.executor
						if dataset.mask is not None:
							logger.debug("selection: %r", dataset.mask.sum())
						if 'state' in arguments:
							dataset.state_set(arguments['state'])
						if expressions:
							for expression in expressions:
								try:
									dataset.validate_expression(expression)
								except (SyntaxError, KeyError, NameError) as e:
									return exception(e)
						subspace = dataset(*expressions, executor=webserver.thread_local.executor) if expressions else None
						if dataset_original.has_selection(vaex.dataset.FILTER_SELECTION_NAME):
							selection_original = dataset_original.get_selection(vaex.dataset.FILTER_SELECTION_NAME)
							if not dataset.has_selection(vaex.dataset.FILTER_SELECTION_NAME):
								dataset.set_selection(selection_original, name=vaex.dataset.FILTER_SELECTION_NAME)
							else:
								selection = dataset.get_selection(vaex.dataset.FILTER_SELECTION_NAME)
								# UGLY: we 'stitch' the current selection at the end of the
								# selection 'list'
								selection_end = selection
								while selection_end.previous_selection:
									selection_end = selection_end.previous_selection
								selection_end.previous_selection = selection_original
								selection_end.mode = "and"
						try:
							if subspace:
								if 'selection' in arguments and arguments['selection']:
									subspace = subspace.selected()
							if subspace is None:
								for name in "job_id state auto_fraction expressions active_fraction selections variables virtual_columns active_start_index active_end_index".split():
									arguments.pop(name, None)
							else:
								if subspace is not None:
									for name in "job_id state auto_fraction expressions active_fraction selection selections variables virtual_columns active_start_index active_end_index".split():
										arguments.pop(name, None)
							logger.debug("subspace: %r", subspace)
							if subspace is None and method_name in "count cov correlation covariance mean std minmax min max sum var".split():
								grid = task_invoke(dataset, method_name, **arguments)
								return grid
							elif method_name in ["minmax", "image_rgba_url", "var", "mean", "sum", "limits_sigma", "nearest", "correlation", "mutual_information"]:
								#print "expressions", expressions
								values = task_invoke(subspace, method_name, **arguments)
								logger.debug("result: %r", values)
								#print values, expressions
								values = values.tolist() if hasattr(values, "tolist") else values

								#for value in values:
								#	logger.debug("value: %r" % type(value))
								#try:
								#	values = [(k.tolist() if hasattr(k, "tolist") else k) for k in values]
								#except:
								#	pass
								return ({"result": values})
							elif method_name == "histogram":
								grid = task_invoke(subspace, method_name, **arguments)
								return grid
								#self.set_header("Content-Type", "application/octet-stream")
								#return (grid.tostring())
								#import base64
								#return grid.tostring()
								#result = base64.b64encode(grid.tostring()).decode("ascii")
								#logger.debug("type: %s" % type(result))
								#return {"result": result}
							elif method_name in ["select", "lasso_select"]:
								dataset.mask = webserver.cache_selection.get((dataset.path, user_id))
								result = task_invoke(dataset, method_name, **arguments)
								result = result.tolist() if hasattr(result, "tolist") else result
								webserver.cache_selection[(dataset.path, user_id)] = dataset.mask
								return ({"result": result})
							elif method_name in ["evaluate"]:
								result = task_invoke(dataset, method_name, **arguments)
								return result
							else:
								logger.error("unknown method: %r", method_name)
								return error("unknown method: " + method_name)
						except (SyntaxError, KeyError, NameError) as e:
							return exception(e)
	except Exception as e:
		logger.exception("unknown issue")
		return exception(e)
	return error("unknown request")

from cachetools import Cache, LRUCache
import sys
MB = 1024**2
GB = MB * 1024

class WebServer(threading.Thread):
	def __init__(self, address="localhost", port=9000, webserver_thread_count=2, cache_byte_size=500*MB,
				 cache_selection_byte_size=500*MB, datasets=[], compress=True, development=False, threads_per_job=4):
		threading.Thread.__init__(self)
		self.setDaemon(True)
		self.address = address
		self.port = port
		self.started = threading.Event()
		self.set_datasets(datasets)

		self.webserver_thread_count = webserver_thread_count
		self.threads_per_job = threads_per_job

		self.thread_pool = concurrent.futures.ThreadPoolExecutor(self.webserver_thread_count)
		self.thread_local = threading.local()
		self.thread_pools = []

		self.job_queue = JobQueue()

		self.cache = LRUCache(cache_byte_size, getsizeof=sys.getsizeof)
		self.cache_selection = LRUCache(cache_selection_byte_size, getsizeof=sys.getsizeof)

		self.options = dict(webserver=self, datasets=datasets, submit_threaded=self.submit_threaded, cache=self.cache,
							cache_selection=self.cache_selection)


		#tornado.web.GZipContentEncoding.MIN_LENGTH = 1
		tornado.web.GZipContentEncoding.CONTENT_TYPES.add("application/octet-stream")
		self.application = tornado.web.Application([
			(r"/queue", QueueHandler, self.options),
			(r"/auth", GoogleOAuth2LoginHandler, {}),
			(r"/websocket", ProgressWebSocket, self.options),
			(r"/.*", ListHandler, self.options),
		], compress_response=compress, debug=development)
		logger.debug("compression set to %r", compress)
		logger.debug("cache size set to %s", vaex.utils.filesize_format(cache_byte_size))
		logger.debug("thread count set to %r", self.webserver_thread_count)

	def set_datasets(self, datasets):
		self.datasets = list(datasets)
		self.datasets_map = dict([(ds.name,ds) for ds in self.datasets])

	def submit_threaded(self, callable, *args, **kwargs):
		job = JobFlexible(4., callable, args=args, kwargs=kwargs)
		self.job_queue.add(job)
		def execute():
			job = self.job_queue.get_next()
			job.kwargs["fraction"] = job.fraction # add fraction keyword argument to the callback
			try:
				result = job.execute()
			finally:
				self.job_queue.finished(job)
			return result
		future = self.thread_pool.submit(execute) #, *args, **kwargs)
		return future

	def serve(self):
		self.mainloop()

	def serve_threaded(self):
		logger.debug("start thread")
		self.start()
		logger.debug("wait for thread to run")
		self.started.wait()
		logger.debug("make tornado io loop the main thread's current")
		# this will make the main thread use this ioloop as current
		self.ioloop.make_current()

	def run(self):
		self.mainloop()

	def mainloop(self):
		logger.info("serving at http://%s:%d" % (self.address, self.port))
		self.ioloop = tornado.ioloop.IOLoop.current()
		# listen doesn't return a server object, which we need to close
		#self.application.listen(self.port, address=self.address)
		from tornado.httpserver import HTTPServer
		self.server = HTTPServer(self.application)
		try:
			self.server.listen(self.port, self.address)
		except:
			self.started.set()
			raise
		self.started.set()
		#self.ioloop.add_callback(self.started.set)
		#if not self.ioloop.is_running():
		try:
			self.ioloop.start()
		except RuntimeError:
			pass # TODO: not sure why this happens in the unittest
		#self.ioloop.stop()
		#try:
		#	self.ioloop.close()
		#except ValueError:
		#	pass
		#self.ioloop.clear_current()

	def stop_serving(self):
		logger.debug("stop server")
		self.server.stop()
		logger.debug("stop io loop")
		self.ioloop.stop()
		#self.ioloop.close()
		#self.ioloop.clear_current()
		for thread_pool in self.thread_pools:
			thread_pool.shutdown()

defaults_yaml = """
address: 0.0.0.0
port: 9000
filenames: []
verbose: 2
cache: 500000000
compress: true
filename: []
development: False
threads_per_job: 4
"""

def main(argv):

	parser = argparse.ArgumentParser(argv[0])
	parser.add_argument("filename", help="filename for dataset", nargs='*')
	parser.add_argument("--address", help="address to bind the server to (default: %(default)s)", default="0.0.0.0")
	parser.add_argument("--port", help="port to listen on (default: %(default)s)", type=int, default=9000)
	parser.add_argument('--verbose', '-v', action='count', default=2)
	parser.add_argument('--cache', help="cache size in bytes for requests, set to zero to disable (default: %(default)s)", type=int, default=500000000)
	parser.add_argument('--compress', help="compress larger replies (default: %(default)s)", default=True, action='store_true')
	parser.add_argument('--no-compress', dest="compress", action='store_false')
	parser.add_argument('--development', default=False, action='store_true', help="enable development features (auto reloading)")
	parser.add_argument('--threads-per-job', default=4, type=int, help="threads per job (default: %(default)s)")
	#config = layeredconfig.LayeredConfig(defaults, env, layeredconfig.Commandline(parser=parser, commandline=argv[1:]))
	config = parser.parse_args(argv[1:])

	verbosity = ["ERROR", "WARNING", "INFO", "DEBUG"]
	logging.getLogger("vaex").setLevel(verbosity[config.verbose])
	#import vaex
	#vaex.set_log_level_debug()
	from vaex.settings import webserver as settings

	#filenames = config.filenames
	filenames = []
	filenames = config.filename
	datasets = []
	for filename in filenames:
		ds = vx.open(filename)
		if ds is None:
			print("error opening file: %r" % filename)
		else:
			datasets.append(ds)
	datasets = datasets or [vx.example()]
	#datasets = [ds for ds in datasets if ds is not None]
	logger.info("datasets:")
	for dataset in datasets:
		logger.info("\thttp://%s:%d/%s or ws://%s:%d/%s", config.address, config.port, dataset.name, config.address, config.port, dataset.name)
	server = WebServer(datasets=datasets, address=config.address, port=config.port, cache_byte_size=config.cache,
					   compress=config.compress, development=config.development,
					   threads_per_job=config.threads_per_job)
	server.serve()

if __name__ == "__main__":
	#logger.setLevel(logging.logging.DEBUG)
	#print sys.argv
	main(sys.argv)
	#3_threaded()
	#import time
	#time.sleep(10)
	#server.stop_serving()


# -*- coding: utf-8 -*-
#from vaex import dataset
import vaex.dataset
from optparse import OptionParser
from mab.utils.progressbar import ProgressBar
import sys
import h5py
import numpy as np


def merge(output_filename, datasets_list, datasets_centering=None, sort_property=None, order_column_name=None, ascending=True):
	#datasets = list(datasets)
	if sort_property:
		datasets_list.sort(key=lambda datasets: datasets[0].variables[sort_property], reverse=not ascending)
		datasets_centering.sort(key=lambda dataset: dataset.variables[sort_property], reverse=not ascending)

	h5output = h5py.File(output_filename, "w")
	example_dataset = datasets_list[0][0]
	max_length = max([sum(len(dataset) for dataset in datasets) for datasets in datasets_list])
	#counts =

	#for datasets in datasets:
	#max_length = sum( counts )
	shape = (len(datasets_list), max_length)
	print(("shape of new arrays will be", shape, max_length))
	if 0:
		for dataset1 in datasets:
			for dataset2 in datasets:
				if dataset1 != dataset2:
					if len(dataset1) != len(dataset2):
						print((dataset1.name, "is of length", len(dataset1), "but", dataset2.name, "is of length", len(dataset2)))
						sys.exit(1)
		
	for column_name in example_dataset.column_names:
		d = h5output.require_dataset("/columns/"+column_name, shape=shape, dtype=example_dataset.columns[column_name].dtype.type, exact=True)
		d[0,0] = example_dataset.columns[column_name][0] # ensure the array exists
	# each float propery will be a new axis in the merged file (TODO: int and other types?)
	for property_name in list(example_dataset.variables.keys()):
		property = example_dataset.variables[property_name]
		if isinstance(property, float):
			d = h5output.require_dataset("/axes/"+property_name, shape=(len(datasets_list),), dtype=np.float64, exact=True)
			d[0] = 0. # make sure it exists
	# close file and open it again with our interface
	h5output.close()
	dataset_output = vaex.dataset.Hdf5MemoryMapped(output_filename, write=True)
		
	progressBar = ProgressBar(0, len(datasets_list)-1)

	if 0:
		idmap = {}
		for index, dataset in enumerate(datasets_list):
			ids = dataset.columns["ParticleIDs"]
			for id in ids:
				idmap[id] = None
		used_ids = list(idmap.keys())
		print((sorted(used_ids)))

	particle_type_count = len(datasets_list[0])
	for index, datasets in enumerate(datasets_list):

		centers = {}
		if datasets_centering is not None:
			cols = datasets_centering[index].columns
			# first rought estimate
			for name in "x y z vx vy vz".split():
				indices = np.argsort(cols["Potential"])[::1]
				indices = indices[:10]
				#centers[name] = cols[name][np.argmin(cols["Potential"])] #.mean()
				centers[name] = cols[name].mean()
				print(("center", centers[name]))

			if 0:
				#if column_name in "x y z".split():

				# now sort by r
				r = np.sqrt(np.sum( [(cols[name]-centers[name])**2 for name in "x y z".split()], axis=0 ))
				indices = np.argsort(r)
				indices = indices[:len(indices)/2] # take 50%

				for name in "x y z".split():
					centers[name] = cols[name][indices].mean()

				# sort by v
				v = np.sqrt(np.sum( [(cols[name]-centers[name])**2 for name in "vx vy vz".split()] ))
				indices = np.argsort(r)
				indices = indices[:len(indices)/2] # take 50%

				for name in "vx vy vz".split():
					centers[name] = cols[name][indices].mean()



		for column_name in datasets[0].column_names:
			column_output = dataset_output.rank1s[column_name]
			column_output[index,:] = np.nan # first fill with nan's since the length of the output column may be larger than that of individual input datasets

			for property_name in list(datasets[0].variables.keys()):
				property = datasets[0].variables[property_name]
				if isinstance(property, float):
					#print "propery ignored: %r" % property
					#print "propery set: %s %r" % (property_name, property)
					dataset_output.axes[property_name][index] = property
				else:
					#print "propery ignored: %s %r" % (property_name, property)
					pass

			center = 0

			output_offset = 0
			# merge the multiple datasets into the one column
			for particle_type_index in range(particle_type_count):
				dataset = datasets[particle_type_index]
				#print len(dataset), output_offset

				column_input = dataset.columns[column_name]
				if order_column_name:
					order_column = dataset.columns[order_column_name]
				else:
					order_column = None
				#print dataset.name, order_column, order_column-order_column.min()
				i1, i2 = output_offset, output_offset + len(dataset)
				if order_column is not None:
					column_output[index,order_column-order_column.min()] = column_input[:] - centers.get(column_name, 0)
				else:
					column_output[index,i1:i1+len(dataset)] = column_input[:] - centers.get(column_name, 0)
				output_offset += len(dataset)
			#print "one file"
			progressBar.update(index)
		
	

	
	
if __name__ == "__main__":
	usage = "use the source luke!"
	parser = OptionParser(usage=usage)

	#parser.add_option("-n", "--name",
	#                 help="dataset name [default=%default]", default="data", type=str)
	parser.add_option("-o", "--order", default=None, help="rows in the input file are ordered by this column (For gadget: ParticleID)")
	parser.add_option("-t", "--type", default=None, help="file type")
	parser.add_option("-p", "--particle-types", default=None, help="gadget particle type")
	parser.add_option("-c", "--center_type", default=None, help="gadget centering type")
	#parser.add_option("-i", "--ignore", default=None, help="ignore errors while loading files")
	parser.add_option("-r", "--reverse", action="store_true", default=False, help="reverse sorting")
	parser.add_option("-s", "--sort",
	                 help="sort datasets by propery [by default it will be the file order]", default=None, type=str)
	(options, args) = parser.parse_args()
	inputs = args[:-1]
	output = args[-1]
	print(("merging:", "\n\t".join(inputs)))
	print(("to:", output))
	if options.type is None:
		print("specify file type --type")
		parser.print_help()
		sys.exit(1)
	#dataset_type_and_options = options.format.split(":")
	#dataset_type, dataset_options = dataset_type_and_options[0], dataset_type_and_options[1:]
	#if dataset_type not in vaex.dataset.dataset_type_map:
	#	print "unknown type", dataset_type
	#	print "possible options are:\n\t", "\n\t".join(vaex.dataset.dataset_type_map.keys())
	#	sys.exit(1)
	#evaluated_options = []
	#f#or option in dataset_options:
	#	evaluated_options.append(eval(option))
		
	class_ = vaex.dataset.dataset_type_map[options.type]
	#@datasets = [class_(filename, *options) for filename in inputs]
	datasets_list = []
	for filename in inputs:
		print(("loading file", filename,options.particle_types))
		datasets = []
		for type in options.particle_types.split(","):
			datasets.append(class_(filename + "#" + type))
		datasets_list.append(datasets)
	datasets_centering = None
	if options.center_type:
		datasets_centering = []
		for filename in inputs:
			datasets_centering.append(class_(filename + "#" + options.center_type))

	merge(output, datasets_list, datasets_centering, options.sort, ascending=not options.reverse, order_column_name=options.order)
	
	